# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains builder and controller class definitions for the quantization algorithm.
"""
from collections import Counter
from collections import OrderedDict
from copy import deepcopy
from enum import IntEnum
from string import Template
from typing import Any, Dict, List, Optional, Set, Tuple

import torch

import nncf
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.common.deprecation import warning_deprecated
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.definitions import MODEL_INPUT_OP_NAME
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import WeightedLayerAttributes
from nncf.common.graph.patterns.manager import PatternsManager
from nncf.common.graph.patterns.manager import TargetDevice
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.utils import get_first_nodes_of_type
from nncf.common.hardware.config import HWConfig
from nncf.common.hardware.config import HWConfigType
from nncf.common.hardware.config import get_hw_config_type
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.quantization.config_assignment import assign_qconfig_lists_to_modules
from nncf.common.quantization.quantizer_propagation.structs import IgnoreReason
from nncf.common.quantization.quantizer_setup import DEFAULT_QUANTIZER_CONFIG
from nncf.common.quantization.quantizer_setup import MultiConfigQuantizerSetup
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import QuantizerSetupBase
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.common.quantization.structs import QuantizableWeightedLayerNode
from nncf.common.quantization.structs import QuantizationConstraints
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.quantization.structs import QuantizerGroup
from nncf.common.quantization.structs import QuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import copy_model
from nncf.common.utils.debug import is_debug
from nncf.config import NNCFConfig
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.config.extractors import extract_range_init_params
from nncf.config.schemata.algo.quantization import PRECISION_INIT_TYPES_VS_DESCRIPTION
from nncf.config.schemata.defaults import QUANTIZATION_EXPORT_TO_ONNX_STANDARD_OPS
from nncf.config.schemata.defaults import QUANTIZATION_LOGARITHM_SCALE
from nncf.config.schemata.defaults import QUANTIZATION_OVERFLOW_FIX
from nncf.config.schemata.defaults import QUANTIZATION_PRESET
from nncf.config.schemata.defaults import QUANTIZE_INPUTS
from nncf.config.schemata.defaults import QUANTIZE_OUTPUTS
from nncf.experimental.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import TensorStatistic
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.algo_selector import ZeroCompressionLoss
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import ELEMENTWISE_OPERATIONS
from nncf.torch.graph.operator_metatypes import UNIFICATION_PRODUCING_METATYPES
from nncf.torch.graph.operator_metatypes import PTCatMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv2dSubtype
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TransformationPriority
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.hardware.config import PTHWConfig
from nncf.torch.initialization import SimpleDataLoaderRunner
from nncf.torch.module_operations import UpdatePaddingValue
from nncf.torch.nncf_network import LoadStateListener
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.adjust_padding import AdjustPaddingArgs
from nncf.torch.quantization.adjust_padding import CalculatePaddingAdjustment
from nncf.torch.quantization.base_ctrl import QuantizationControllerBase
from nncf.torch.quantization.debug_interface import QuantizationDebugInterface
from nncf.torch.quantization.default_quantization import DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT
from nncf.torch.quantization.default_quantization import QUANTIZATION_LAYER_METATYPES
from nncf.torch.quantization.external_quantizer import ExternalQuantizerCallHook
from nncf.torch.quantization.init_precision import PrecisionInitializerFactory
from nncf.torch.quantization.init_range import DataLoaderRangeInitializeRunner
from nncf.torch.quantization.init_range import PTRangeInitParams
from nncf.torch.quantization.init_range import StatCollectorGenerator
from nncf.torch.quantization.layers import QUANTIZATION_MODULES
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import PTQuantizationPoint
from nncf.torch.quantization.layers import PTQuantizerSetup
from nncf.torch.quantization.layers import PTQuantizerSpec
from nncf.torch.quantization.layers import QuantizerConfig
from nncf.torch.quantization.layers import QuantizerExportMode
from nncf.torch.quantization.layers import QuantizersSwitcher
from nncf.torch.quantization.layers import SymmetricQuantizer
from nncf.torch.quantization.layers import get_scale_shape
from nncf.torch.quantization.metrics import MemoryConsumptionStatisticsCollector
from nncf.torch.quantization.metrics import PTQuantizationStatisticsCollector
from nncf.torch.quantization.metrics import QuantizationShareBuildTimeInfo
from nncf.torch.quantization.metrics import ShareEdgesQuantizedDataPathStatisticsCollector
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.precision_init.adjacent_quantizers import GroupsOfAdjacentQuantizers
from nncf.torch.quantization.precision_init.autoq_init import AutoQPrecisionInitParams
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitializer
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitParams
from nncf.torch.quantization.precision_init.hawq_init import HAWQPrecisionInitParams
from nncf.torch.quantization.precision_init.manual_init import ManualPrecisionInitParams
from nncf.torch.quantization.schedulers import QUANTIZATION_SCHEDULERS
from nncf.torch.quantization.strip import strip_quantized_model
from nncf.torch.quantization.structs import NonWeightQuantizerInfo
from nncf.torch.quantization.structs import WeightQuantizerInfo
from nncf.torch.quantization.translator import PTTargetPointTranslator
from nncf.torch.structures import AutoQPrecisionInitArgs
from nncf.torch.structures import QuantizationPrecisionInitArgs
from nncf.torch.tensor_statistics.algo import TensorStatisticsCollectionBuilder
from nncf.torch.tensor_statistics.statistics import pt_convert_stat_to_min_max_tensor_stat
from nncf.torch.utils import get_model_device
from nncf.torch.utils import get_model_dtype
from nncf.torch.utils import get_state_dict_names_with_modules
from nncf.torch.utils import is_main_process
from nncf.torch.utils import training_mode_switcher

QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME = "version"


class QuantizerBuilderStateVersion(IntEnum):
    # In Quantization builder state SingleConfigQuantizerSetup is being saved as quantizer setup.
    v1 = 1
    # In Quantization builder state PTQuantizerSetup is being saved as quantizer setup.
    v2 = 2

    @staticmethod
    def from_compression_state(compression_state):
        if QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME in compression_state:
            return compression_state.get(QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME)
        return QuantizerBuilderStateVersion.v1


class QuantizerSetupGeneratorBase:
    def __init__(
        self,
        quant_config: Dict,
        target_model: NNCFNetwork,
        precision_init_type: str = None,
        precision_init_params: BasePrecisionInitParams = None,
        range_init_params: PTRangeInitParams = None,
        hw_config: HWConfig = None,
    ):
        self._target_model = target_model
        self._quantization_config = quant_config
        self.hw_config = hw_config
        self._target_device = None if hw_config is None else hw_config.target_device
        self._quantize_inputs = self._quantization_config.get("quantize_inputs", QUANTIZE_INPUTS)
        self._quantize_outputs = self._quantization_config.get("quantize_outputs", QUANTIZE_OUTPUTS)

        self.ignored_scopes = self._quantization_config.get("ignored_scopes")
        self.target_scopes = self._quantization_config.get("target_scopes")

        self.global_quantizer_constraints: Dict[QuantizerGroup, QuantizationConstraints] = {}
        self._ignored_scopes_per_group: Dict[QuantizerGroup, List[str]] = {}
        self._target_scopes_per_group: Dict[QuantizerGroup, List[str]] = {}

        for quantizer_group in QuantizerGroup:
            self._parse_group_params(self._quantization_config, quantizer_group)

        self._precision_init_type = precision_init_type
        self._precision_init_params = precision_init_params
        self._range_init_params = range_init_params
        self._num_potential_quantized_weights = len(self._target_model.nncf.get_nncf_modules())

    def generate_setup(self) -> SingleConfigQuantizerSetup:
        raise NotImplementedError

    def get_build_time_metric_infos(self):
        raise NotImplementedError

    def _parse_group_params(self, quant_config: Dict, quantizer_group: QuantizerGroup):
        group_name = quantizer_group.value
        params_dict = {}
        params_dict_from_config = quant_config.get(group_name, {})
        preset = quant_config.get("preset")
        if self._target_device in ["ANY", "CPU", "GPU"] or (self._target_device is None and preset is not None):
            preset = QuantizationPreset(quant_config.get("preset", QUANTIZATION_PRESET))
            params_dict = preset.get_params_configured_by_preset(quantizer_group)
            overridden_params = params_dict.keys() & params_dict_from_config.keys()
            if overridden_params:
                nncf_logger.info(f"Preset quantizer parameters {overridden_params} explicitly overridden by config.")
        params_dict.update(params_dict_from_config)
        self.global_quantizer_constraints[quantizer_group] = QuantizationConstraints.from_config_dict(params_dict)
        self._ignored_scopes_per_group[quantizer_group] = params_dict_from_config.get("ignored_scopes", [])
        if self.ignored_scopes is not None:
            self._ignored_scopes_per_group[quantizer_group] += self.ignored_scopes
        target_scopes = params_dict_from_config.get("target_scopes")
        if target_scopes is None and self.target_scopes is not None:
            self._target_scopes_per_group[quantizer_group] = self.target_scopes
        else:
            self._target_scopes_per_group[quantizer_group] = target_scopes

    def _get_default_qconfig(self, constraints: QuantizationConstraints = None):
        qconfig = deepcopy(DEFAULT_QUANTIZER_CONFIG)
        if constraints is not None:
            qconfig = constraints.apply_constraints_to(qconfig)
        return qconfig

    def _filter_by_ignored_algo(self, nodes: List[NNCFNode]) -> List[NNCFNode]:
        retval = []
        for node in nodes:
            if "quantization" in node.ignored_algorithms:
                continue
            retval.append(node)
        return retval

    def _assign_qconfig_lists_to_modules(self, weighted_nodes: List[NNCFNode]) -> Dict[NNCFNode, List[QuantizerConfig]]:
        raise NotImplementedError

    def get_quantizable_module_nodes(self) -> List[QuantizableWeightedLayerNode]:
        weighted_nodes = self._target_model.nncf.get_original_graph().get_nodes_by_metatypes(
            QUANTIZATION_LAYER_METATYPES
        )
        quantized_modules_with_potential_qconfig = []

        weighted_nodes = self._filter_by_ignored_algo(weighted_nodes)
        weighted_node_vs_qconfig_list = self._assign_qconfig_lists_to_modules(weighted_nodes)

        for node, qconfig_list in weighted_node_vs_qconfig_list.items():
            if qconfig_list is not None:
                qconfig_list_copy = deepcopy(qconfig_list)
                quantized_modules_with_potential_qconfig.append(QuantizableWeightedLayerNode(node, qconfig_list_copy))
        return quantized_modules_with_potential_qconfig


class IQuantizerSetupDisambiguator:
    def select_final_quantizer_setup(self, multi_config_setup: MultiConfigQuantizerSetup) -> SingleConfigQuantizerSetup:
        raise NotImplementedError


class DefaultQuantizerSetupDisambiguator(IQuantizerSetupDisambiguator):
    def __init__(
        self,
        target_model: NNCFNetwork,
        precision_init_type: str = None,
        precision_init_params: BasePrecisionInitParams = None,
        range_init_params: PTRangeInitParams = None,
        override_bit_options_with_precision_init: bool = False,
        hw_config: HWConfig = None,
    ):
        self._precision_init_type = precision_init_type
        self._precision_init_params = precision_init_params
        self._range_init_params = range_init_params
        self._target_model = target_model
        self._override_bit_options_with_precision_init = override_bit_options_with_precision_init
        self.hw_config = hw_config

    @staticmethod
    def select_first_qconfig_with_bitwidth_variants_for_each_point(
        multi_config_setup: MultiConfigQuantizerSetup,
    ) -> MultiConfigQuantizerSetup:
        new_setup = deepcopy(multi_config_setup)
        for qp_id, qp in multi_config_setup.quantization_points.items():
            main_qconfig = qp.possible_qconfigs[0]
            constrained_qconfig_list = [main_qconfig]
            if len(qp.possible_qconfigs) > 1:
                constrained_qconfig_list += list(filter(main_qconfig.is_a_bitwidth_variant, qp.possible_qconfigs[1:]))
            new_setup.quantization_points[qp_id].possible_qconfigs = constrained_qconfig_list
        return new_setup

    def select_final_quantizer_setup(self, multi_config_setup: MultiConfigQuantizerSetup) -> SingleConfigQuantizerSetup:
        if self._precision_init_type is not None:
            with self._target_model.nncf.temporary_clean_view() as intermediate_model:
                stats = QuantizationBuilder.get_statistics_for_quantizer_setup(
                    intermediate_model, multi_config_setup, self._range_init_params
                )
                bitwidth_varying_only_multi_setup = self.select_first_qconfig_with_bitwidth_variants_for_each_point(
                    multi_config_setup
                )

                init_setup = bitwidth_varying_only_multi_setup.select_first_qconfig_for_each_point()
                intermediate_builder = ExperimentalQuantizationBuilder(
                    bitwidth_varying_only_multi_setup, init_setup, stats, hw_config=self.hw_config
                )
                intermediate_builder.apply_to(intermediate_model)
                intermediate_ctrl = intermediate_builder.build_controller(intermediate_model)

                # intermediate_ctrl.init_range()
                hw_constraints = HardwareQuantizationConstraints()
                if not self._override_bit_options_with_precision_init:
                    for qp_id, qp in multi_config_setup.quantization_points.items():
                        quantizer_module_id = intermediate_ctrl.setup_to_module_id_translation_dict[qp_id]
                        hw_constraints.add(quantizer_module_id, qp.possible_qconfigs)
                final_quantizer_setup = intermediate_ctrl.init_precision(
                    self._precision_init_type, self._precision_init_params, hw_constraints
                )
        else:
            final_quantizer_setup = multi_config_setup.select_first_qconfig_for_each_point()
        return final_quantizer_setup


class PropagationBasedQuantizerSetupGenerator(QuantizerSetupGeneratorBase):
    def __init__(
        self,
        quant_config: Dict,
        target_model: NNCFNetwork,
        hw_config: HWConfig = None,
        device: TargetDevice = None,
        precision_init_type: str = None,
        precision_init_params: BasePrecisionInitParams = None,
        range_init_params: PTRangeInitParams = None,
        debug_interface: "QuantizationDebugInterface" = None,
    ):
        super().__init__(
            quant_config, target_model, precision_init_type, precision_init_params, range_init_params, hw_config
        )

        self._pattern_fusing_graph = PatternsManager.get_full_hw_pattern_graph(backend=BackendType.TORCH, device=device)

        self._hw_precision_constraints = HardwareQuantizationConstraints()
        self._debug_interface = debug_interface
        self._num_potential_quantized_activations = 0

        act_config = quant_config.get(QuantizerGroup.ACTIVATIONS.value, {})
        self._unified_scale_ops = act_config.get("unified_scale_ops")

    def generate_setup(self) -> SingleConfigQuantizerSetup:
        quantizable_module_nodes = self.get_quantizable_module_nodes()

        insertion_point_graph = self._target_model.nncf.get_original_insertion_point_graph()
        if self._debug_interface:
            self._debug_interface.visualize_insertion_point_graph(insertion_point_graph)
        from nncf.common.quantization.quantizer_propagation.solver import QuantizerPropagationSolver

        scales_unification_map = {PTCatMetatype: UNIFICATION_PRODUCING_METATYPES}
        ignored_scopes_for_solver = {
            name: IgnoreReason.USER_REQUESTED for name in self._ignored_scopes_per_group[QuantizerGroup.ACTIVATIONS]
        }
        prop_graph_solver = QuantizerPropagationSolver(
            activation_ignored_scopes=ignored_scopes_for_solver,
            weight_ignored_scopes=self._ignored_scopes_per_group[QuantizerGroup.WEIGHTS],
            activation_target_scopes=self._target_scopes_per_group[QuantizerGroup.ACTIVATIONS],
            weight_target_scopes=self._target_scopes_per_group[QuantizerGroup.WEIGHTS],
            hw_config=self.hw_config,
            default_trait_to_metatype_map=DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT,
            default_qconfig_list=[
                self._get_default_qconfig(constraints=self.global_quantizer_constraints[QuantizerGroup.ACTIVATIONS])
            ],
            quantizable_layer_nodes=quantizable_module_nodes,
            scope_overrides=self._quantization_config.get("scope_overrides", {}),
            global_constraints=self.global_quantizer_constraints,
            additional_unified_scale_op_scopes=self._unified_scale_ops,
            quantize_outputs=self._quantize_outputs,
            scales_unification_map=scales_unification_map,
        )

        merged_ip_graph = insertion_point_graph.get_ip_graph_with_merged_hw_optimized_operations(
            self._pattern_fusing_graph
        )
        quantization_proposal = prop_graph_solver.run_on_ip_graph(merged_ip_graph, ELEMENTWISE_OPERATIONS)
        self._num_potential_quantized_activations = prop_graph_solver.get_num_potential_quantized_activations()

        quantizer_setup = deepcopy(quantization_proposal.quantizer_setup)
        quantization_proposal.quantizer_setup = quantizer_setup

        disambiguator = DefaultQuantizerSetupDisambiguator(
            self._target_model,
            self._precision_init_type,
            self._precision_init_params,
            self._range_init_params,
            override_bit_options_with_precision_init=self.hw_config is None,
            hw_config=self.hw_config,
        )

        single_config_quantizer_setup = disambiguator.select_final_quantizer_setup(
            quantization_proposal.quantizer_setup
        )

        finalized_proposal = quantization_proposal.finalize(
            single_config_quantizer_setup, strict=self.hw_config is not None
        )
        finalized_quantizer_setup = prop_graph_solver.get_final_quantizer_setup(finalized_proposal)
        finalized_quantizer_setup = self._handle_quantize_inputs_option(finalized_quantizer_setup)
        return finalized_quantizer_setup

    def _assign_qconfig_lists_to_modules(self, weighted_nodes: List[NNCFNode]) -> Dict[NNCFNode, List[QuantizerConfig]]:
        global_constraints = self.global_quantizer_constraints[QuantizerGroup.WEIGHTS]
        scope_overrides_dict = self._quantization_config.get("scope_overrides", {})
        return assign_qconfig_lists_to_modules(
            weighted_nodes, self._get_default_qconfig(), global_constraints, scope_overrides_dict, self.hw_config
        )

    def _handle_quantize_inputs_option(self, quantizer_setup: SingleConfigQuantizerSetup) -> SingleConfigQuantizerSetup:
        nncf_graph = self._target_model.nncf.get_original_graph()
        qp_ids_to_discard = []
        for qp_id, qp in quantizer_setup.quantization_points.items():
            if qp.is_activation_quantization_point():
                insertion_point = qp.insertion_point
                target_node = nncf_graph.get_node_by_name(insertion_point.target_node_name)
                if not self._quantize_inputs and target_node.node_type == MODEL_INPUT_OP_NAME:
                    qp_ids_to_discard.append(qp_id)
        for qp_id in qp_ids_to_discard:
            quantizer_setup.discard(qp_id, keep_shared_input_qps=True)
        return quantizer_setup

    def get_build_time_metric_infos(self):
        return QuantizationShareBuildTimeInfo(
            self._num_potential_quantized_activations, self._num_potential_quantized_weights
        )


class QBuilderStateNames:
    BUILD_TIME_METRIC_INFOS = "build_time_metric_infos"
    QUANTIZER_SETUP = "quantizer_setup"


@PT_COMPRESSION_ALGORITHMS.register("quantization")
class QuantizationBuilder(PTCompressionAlgorithmBuilder):
    _state_names = QBuilderStateNames

    def __init__(self, config, should_init: bool = True):
        super().__init__(config, should_init)
        self._debug_interface = QuantizationDebugInterface() if is_debug() else None
        self._weight_quantizers = OrderedDict()  # Quantizers applied via UpdateWeights
        self._non_weight_quantizers = OrderedDict()  # All the other quantizers
        self._quantizers_input_shapes = OrderedDict()
        self._processed_insertion_points: Set[PTTargetPoint] = set()
        self._groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers = GroupsOfAdjacentQuantizers()
        self._setup_to_module_id_translation_dict: Dict[QuantizationPointId, QuantizerId] = {}
        self.eval_ops_exec_ctx = []
        self._build_time_metric_infos: Optional[QuantizationShareBuildTimeInfo] = None
        self.hw_config = None
        self._legacy_single_config_quantizer_setup_from_comp_state: Optional[SingleConfigQuantizerSetup] = None
        self._pt_quantizer_setup: Optional[PTQuantizerSetup] = None
        self._minmax_values_for_range_init: Optional[Dict[QuantizationPointId, MinMaxTensorStatistic]] = {}

        # can be False to disable setting of adjust padding operations on precision init, because it may add unnecessary
        # noise on model evaluation (e.g. in AutoQ)
        self._should_setup_adjust_pad_ops = True
        hw_config_type = None
        self._target_device = self.config.get("target_device", "ANY")
        hw_config_type = get_hw_config_type(self._target_device)
        if hw_config_type is not None:
            hw_config_path = PTHWConfig.get_path_to_hw_config(hw_config_type)
            self.hw_config = PTHWConfig.from_json(hw_config_path)

        algo_config = self._get_algo_specific_config_section()
        if self._target_device == "NPU" and "preset" in algo_config:
            raise nncf.InternalError("The NPU target device does not support presets.")
        if self._target_device == "CPU_SPR":
            raise nncf.InternalError("The CPU_SPR target device does not supported.")

        self._range_init_params = None
        self._precision_init_type = None
        self._precision_init_params = None
        if self.should_init:
            self._parse_init_params()

        self._use_logarithm_scale_per_group: Dict[QuantizerGroup, bool] = {}

        for quantizer_group in QuantizerGroup:
            group_name = quantizer_group.value
            params_dict = self._algo_config.get(group_name, {})
            self._use_logarithm_scale_per_group[quantizer_group] = params_dict.get(
                "logarithm_scale", QUANTIZATION_LOGARITHM_SCALE
            )

        self._overflow_fix = self._algo_config.get("overflow_fix", QUANTIZATION_OVERFLOW_FIX)
        self._device_for_callable_obj_creation = "cpu"

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Initializes object from the state.

        :param state_without_name: Output of `get_state()` method.
        """
        quantizer_setup_state = state_without_name[self._state_names.QUANTIZER_SETUP]
        version = state_without_name.get(QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME, QuantizerBuilderStateVersion.v1)
        if version == QuantizerBuilderStateVersion.v1:
            self._legacy_single_config_quantizer_setup_from_comp_state = SingleConfigQuantizerSetup.from_state(
                quantizer_setup_state
            )
        else:
            self._pt_quantizer_setup = PTQuantizerSetup.from_state(quantizer_setup_state)
        self._build_time_metric_infos = QuantizationShareBuildTimeInfo.from_state(
            state_without_name[self._state_names.BUILD_TIME_METRIC_INFOS]
        )

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        build_time_metric_infos_state = {}
        if self._build_time_metric_infos:
            build_time_metric_infos_state = self._build_time_metric_infos.get_state()
        quantizer_setup_state = {}
        if self._pt_quantizer_setup:
            quantizer_setup_state = self._pt_quantizer_setup.get_state()
        return {
            self._state_names.QUANTIZER_SETUP: quantizer_setup_state,
            self._state_names.BUILD_TIME_METRIC_INFOS: build_time_metric_infos_state,
            QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME: max(QuantizerBuilderStateVersion).value,
        }

    def _parse_init_params(self):
        self._range_init_params = self._parse_range_init_params()
        self._precision_init_type, self._precision_init_params = self._parse_precision_init_params(
            self._algo_config.get("initializer", {})
        )

    def _parse_range_init_params(self) -> Optional[PTRangeInitParams]:
        range_init_params = extract_range_init_params(self.config)
        return PTRangeInitParams(**range_init_params) if range_init_params is not None else None

    def _parse_precision_init_params(self, initializer_config: Dict) -> Tuple[str, BasePrecisionInitParams]:
        init_precision_config = initializer_config.get("precision")
        if not init_precision_config:
            return None, None
        precision_init_type = init_precision_config.get("type", "manual")
        if precision_init_type not in PRECISION_INIT_TYPES_VS_DESCRIPTION:
            raise nncf.InternalError(f"Unrecognized precision init type: {precision_init_type}")
        if precision_init_type == "hawq":
            try:
                precision_init_args = self.config.get_extra_struct(QuantizationPrecisionInitArgs)
            except KeyError as e:
                raise ValueError(
                    "Specified non-manual precision initialization in the NNCF config, "
                    "but the initializing data loader and loss criterion are not provided as an extra struct. "
                    "Refer to `NNCFConfig.register_extra_structs` and the `QuantizationPrecisionInitArgs` "
                    "class"
                ) from e
            precision_init_params = HAWQPrecisionInitParams.from_config(init_precision_config, precision_init_args)
        elif precision_init_type == "autoq":
            if self.hw_config is not None and self.hw_config.target_device != HWConfigType.NPU.value:
                raise ValueError(
                    "Unsupported device ({}). Automatic Precision Initialization only supports for "
                    "target_device NONE or NPU".format(self.hw_config.target_device)
                )
            try:
                precision_init_args = self.config.get_extra_struct(AutoQPrecisionInitArgs)
            except KeyError as e:
                raise ValueError(
                    "Specified Automated precision initialization in the NNCF config, "
                    "but the initializing data loader and loss criterion are not provided as an extra "
                    "struct. Refer to `NNCFConfig.register_extra_structs` and the "
                    "`AutoQPrecisionInitArgs` class"
                ) from e

            hw_config_type = None
            if self.hw_config is not None:
                hw_config_type = HWConfigType(self.hw_config.target_device)
            precision_init_params = AutoQPrecisionInitParams.from_config(
                init_precision_config, precision_init_args, hw_config_type
            )
        elif precision_init_type == "manual":
            precision_init_params = ManualPrecisionInitParams.from_config(init_precision_config)
        else:
            raise ValueError(f"Unhandled precision init type: {precision_init_type}")
        return precision_init_type, precision_init_params

    def _get_minmax_values_for_quantizer_locations(
        self,
        quantizer_setup: SingleConfigQuantizerSetup,
        tensor_statistics: Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]],
        target_model_graph: PTNNCFGraph,
    ) -> Dict[QuantizationPointId, MinMaxTensorStatistic]:
        retval = {}
        for qp_id, qp in quantizer_setup.quantization_points.items():
            qip = qp.insertion_point
            tp = PTTargetPointTranslator.translate(qip)
            if tp not in tensor_statistics:
                nncf_logger.debug(f"TP {tp} not found in tensor statistics")
                retval[qp_id] = None
            else:
                target_node = target_model_graph.get_node_by_name(tp.target_node_name)
                if qp.is_weight_quantization_point():
                    layer_attrs = target_node.layer_attributes
                    assert isinstance(layer_attrs, WeightedLayerAttributes)
                    input_shape = layer_attrs.get_weight_shape()
                    channel_idx = layer_attrs.get_target_dim_for_compression()
                else:
                    input_shape = target_model_graph.get_input_shape_for_insertion_point(qp.insertion_point)
                    channel_idx = 1  # channel dim for activations
                scale_shape = tuple(
                    get_scale_shape(input_shape, qp.is_weight_quantization_point(), qp.qconfig.per_channel, channel_idx)
                )

                if scale_shape not in tensor_statistics[tp]:
                    nncf_logger.debug(f"Did not collect tensor statistics at {tp} for shape {scale_shape}")
                    retval[qp_id] = None
                else:
                    minmax_stat = pt_convert_stat_to_min_max_tensor_stat(tensor_statistics[tp][scale_shape])
                    retval[qp_id] = minmax_stat
        return retval

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        # TODO (vshampor): a simpler solution would be to always create callables on CPU and
        # to move these to model-specific device upon actual application, but would this impact
        # the time required to create a compressed model?
        self._device_for_callable_obj_creation = get_model_device(target_model)
        target_model.nncf.register_compression_module_type(ExtraCompressionModuleType.EXTERNAL_QUANTIZER)
        if self._pt_quantizer_setup is None:
            self._pt_quantizer_setup = self._get_quantizer_setup(target_model)

        (
            insertion_commands,
            setup_to_module_id_translation_dict,
        ) = self._build_insertion_commands_list_for_quantizer_setup(
            self._pt_quantizer_setup, target_model, self._minmax_values_for_range_init
        )

        transformation_layout = PTTransformationLayout()
        for command in insertion_commands:
            transformation_layout.register(command)

        self._setup_to_module_id_translation_dict = setup_to_module_id_translation_dict
        all_quantizations = {}
        all_quantizations.update({k: v.quantizer_module_ref for k, v in self._weight_quantizers.items()})
        all_quantizations.update({k: v.quantizer_module_ref for k, v in self._non_weight_quantizers.items()})
        self._groups_of_adjacent_quantizers.parse_from_quantizer_setup(
            all_quantizations, self._pt_quantizer_setup, setup_to_module_id_translation_dict
        )

        # NOTE: Order of activations must be the same to correctly broadcast parameters (e.g. scales) in distributed
        # mode (see call of `_dist_broadcast_coalesced` in torch/nn/parallel/distributed.py for more details)

        target_model.nncf.sort_compression_modules(ExtraCompressionModuleType.EXTERNAL_QUANTIZER)

        if self._debug_interface is not None:
            target_model.nncf.debug_interface.add_interface(self._debug_interface)

        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        all_quantizations = get_state_dict_names_with_modules(target_model, quantization_types)
        target_model.nncf._load_listener = LoadStateListener(target_model, all_quantizations)

        return transformation_layout

    @staticmethod
    def get_statistics_for_quantizer_setup(
        target_model: NNCFNetwork, quantizer_setup: QuantizerSetupBase, range_init_params: PTRangeInitParams
    ) -> Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]]:
        if range_init_params is None:
            return {}
        observation_points_vs_collectors_dict = (
            StatCollectorGenerator.generate_collectors_for_range_init_statistics_collection(
                target_model.nncf.get_original_graph(), quantizer_setup, range_init_params
            )
        )

        with target_model.nncf.temporary_clean_view() as intermediate_model:
            stat_builder = TensorStatisticsCollectionBuilder(NNCFConfig(), observation_points_vs_collectors_dict)
            stat_builder.apply_to(intermediate_model)
            stat_ctrl = stat_builder.build_controller(intermediate_model)
            runner = SimpleDataLoaderRunner(intermediate_model, range_init_params.device)
            runner.progressbar_description = "Collecting tensor statistics"
            with training_mode_switcher(intermediate_model, is_training=False):
                # Run statistics collection in eval mode, otherwise it may fail because graph was built in eval mode
                runner.run(range_init_params.init_range_data_loader, range_init_params.get_max_num_init_steps())

        retval = {}
        for ip, rs_vs_collector in stat_ctrl.ip_vs_collector_dict.items():
            retval[ip] = {rs: collector.get_statistics() for rs, collector in rs_vs_collector.items()}
        return retval

    def _get_statistics_for_final_range_init(
        self, target_model: NNCFNetwork, quantizer_setup: QuantizerSetupBase, range_init_params: PTRangeInitParams
    ) -> Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]]:
        return self.get_statistics_for_quantizer_setup(target_model, quantizer_setup, range_init_params)

    def _get_single_config_quantizer_setup(self, target_model) -> SingleConfigQuantizerSetup:
        setup_generator = PropagationBasedQuantizerSetupGenerator(
            self._algo_config,
            target_model,
            self.hw_config,
            self._target_device,
            self._precision_init_type,
            self._precision_init_params,
            self._range_init_params,
            self._debug_interface,
        )
        single_config_quantizer_setup = setup_generator.generate_setup()
        self._build_time_metric_infos = setup_generator.get_build_time_metric_infos()
        return single_config_quantizer_setup

    def _get_quantizer_setup(self, target_model: NNCFNetwork) -> PTQuantizerSetup:
        if self._legacy_single_config_quantizer_setup_from_comp_state is None:
            single_config_quantizer_setup = self._get_single_config_quantizer_setup(target_model)
        else:
            single_config_quantizer_setup = self._legacy_single_config_quantizer_setup_from_comp_state

        target_model_graph = target_model.nncf.get_original_graph()

        if is_main_process() and self.should_init:
            stats_for_range_init = self._get_statistics_for_final_range_init(
                target_model, single_config_quantizer_setup, self._range_init_params
            )
            self._minmax_values_for_range_init = self._get_minmax_values_for_quantizer_locations(
                single_config_quantizer_setup, stats_for_range_init, target_model_graph
            )

            self._check_and_log_missing_stats_for_setup(
                single_config_quantizer_setup, self._minmax_values_for_range_init
            )

        bitwidth_per_scope = BasePrecisionInitializer.get_bitwidth_per_scope(single_config_quantizer_setup)
        str_bw = [str(element) for element in bitwidth_per_scope]
        nncf_logger.debug("\n".join(['\n"bitwidth_per_scope": [', ",\n".join(str_bw), "]"]))

        setup = PTQuantizerSetup(
            single_config_quantizer_setup.unified_scale_groups,
            single_config_quantizer_setup.shared_input_operation_set_groups,
        )

        for qp_id, qp in single_config_quantizer_setup.quantization_points.items():
            qconfig = qp.qconfig
            insertion_point = qp.insertion_point  # QuantizationInsertionPointBase

            compression_lr_multiplier = self._get_compression_lr_multiplier()

            half_range = False
            if self.hw_config and qp.is_weight_quantization_point():
                target_node = target_model_graph.get_node_by_name(insertion_point.target_node_name)
                if self.hw_config.target_device in ["CPU", "ANY"] and qconfig.num_bits == 8:
                    if self._overflow_fix == "enable":
                        half_range = True
                        quantizers_with_overflow_fix_str = "all weight quantizers"
                    elif self._overflow_fix == "first_layer_only":
                        if target_node in get_first_nodes_of_type(target_model_graph, ["conv2d", "conv3d"]):
                            half_range = True
                            quantizers_with_overflow_fix_str = "first convolution weight quantizers"
                    elif self._overflow_fix != "disable":
                        raise nncf.InternalError(f"Unknown overflow fix type: {self._overflow_fix}")
                    if half_range:
                        nncf_logger.debug(f"Overflow issue fix will be applied to {quantizers_with_overflow_fix_str}")

            if qp.is_weight_quantization_point():
                use_logarithm_scale = self._use_logarithm_scale_per_group[QuantizerGroup.WEIGHTS]
                narrow_range = qconfig.num_bits == 8 and not half_range
            else:
                use_logarithm_scale = self._use_logarithm_scale_per_group[QuantizerGroup.ACTIVATIONS]
                narrow_range = False

            if qp.is_weight_quantization_point():
                target_node = target_model_graph.get_node_by_name(insertion_point.target_node_name)
                layer_attributes = target_node.layer_attributes
                assert isinstance(layer_attributes, WeightedLayerAttributes)
                scale_shape = get_scale_shape(
                    layer_attributes.get_weight_shape(),
                    is_weights=True,
                    per_channel=qconfig.per_channel,
                    channel_idx=layer_attributes.get_target_dim_for_compression(),
                )
            else:
                input_shape = target_model_graph.get_input_shape_for_insertion_point(insertion_point)
                scale_shape = get_scale_shape(list(input_shape), is_weights=False, per_channel=qconfig.per_channel)

            qspec = PTQuantizerSpec.from_config(
                qconfig,
                narrow_range=narrow_range,
                scale_shape=tuple(scale_shape),
                logarithm_scale=use_logarithm_scale,
                half_range=half_range,
                is_quantized_on_export=qp.is_weight_quantization_point(),
                compression_lr_multiplier=compression_lr_multiplier,
            )
            pt_qp = PTQuantizationPoint(
                qspec, PTTargetPointTranslator.translate(insertion_point), qp.directly_quantized_operator_node_names
            )
            setup.add_quantization_point(qp_id, pt_qp)

        return setup

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return QuantizationController(
            model,
            self.config,
            self._debug_interface,
            self._weight_quantizers,
            self._non_weight_quantizers,
            self._groups_of_adjacent_quantizers,
            self._quantizers_input_shapes,
            build_time_metric_info=self._build_time_metric_infos,
            build_time_range_init_params=self._range_init_params,
        )

    def __create_quantize_module(self, quantizer_spec: PTQuantizerSpec):
        quantizer_cls = QUANTIZATION_MODULES.get(quantizer_spec.mode)
        return quantizer_cls(quantizer_spec)

    @staticmethod
    def _get_adjust_padding_args(
        target_model_graph: NNCFGraph,
        quantization_point: PTQuantizationPoint,
        activation_quantizer: BaseQuantizer,
        quantization_points: List[PTQuantizationPoint],
    ) -> List[AdjustPaddingArgs]:
        result = []
        for op_node_name in quantization_point.directly_quantized_operator_node_names:
            weight_bitwidth = None
            for qp in quantization_points:
                is_weight = qp.is_weight_quantization_point()
                if is_weight and (qp.target_point.target_node_name == op_node_name):
                    weight_bitwidth = qp.qspec.num_bits
                    break
            if weight_bitwidth:
                is_applicable = False
                target_node = target_model_graph.get_node_by_name(op_node_name)
                if target_node.metatype in [PTModuleConv2dMetatype, PTModuleDepthwiseConv2dSubtype]:
                    layer_attrs = target_node.layer_attributes
                    assert isinstance(layer_attrs, ConvolutionLayerAttributes)
                    padding_values = set(layer_attrs.padding_values)
                    padding_enabled = len(padding_values) >= 1 and padding_values.pop()
                    if padding_enabled:
                        symmetric = isinstance(activation_quantizer, SymmetricQuantizer)
                        per_tensor = not activation_quantizer.per_channel
                        a_int4 = activation_quantizer.num_bits == 4
                        w_int24 = weight_bitwidth <= 4
                        unsigned = not activation_quantizer.signed
                        is_applicable = symmetric and per_tensor and a_int4 and w_int24 and unsigned
                if is_applicable:
                    result.append(AdjustPaddingArgs(weight_bitwidth, activation_quantizer, op_node_name))
        return result

    def _add_adjust_padding_ops(self, adjust_padding_args: List[AdjustPaddingArgs]):
        commands = []
        for args in adjust_padding_args:
            ap = CalculatePaddingAdjustment(args.activation_quantizer)
            op = UpdatePaddingValue(ap).to(self._device_for_callable_obj_creation)
            insertion_point = PTTargetPoint(
                target_type=TargetType.PRE_LAYER_OPERATION, target_node_name=args.module_op_node_name
            )
            nncf_logger.debug_once(f"Padding will be adjusted for {args.module_op_node_name}")
            commands.append(PTInsertionCommand(insertion_point, op, TransformationPriority.DEFAULT_PRIORITY))
        return commands

    @staticmethod
    def _check_and_log_missing_stats_for_setup(
        quantizer_setup: SingleConfigQuantizerSetup,
        minmax_values_for_range_init: Dict[QuantizationPointId, MinMaxTensorStatistic],
    ):
        tps_with_uncollected_stats = set()
        for qp_id in quantizer_setup.quantization_points:
            if qp_id not in minmax_values_for_range_init:
                tps_with_uncollected_stats.add(quantizer_setup.quantization_points[qp_id].insertion_point)
        if tps_with_uncollected_stats:
            nncf_logger.error("Tensor statistics for the following locations were not collected:")
            for tp in tps_with_uncollected_stats:
                nncf_logger.error(f"\t{tp}")
            nncf_logger.error(
                "The corresponding quantizer range will not be initialized! If the model has "
                "data-dependent control flow branches, make sure that your initializing data loader is "
                "producing data that allows the model cover to all of these branches. If this is not the "
                "case, consider adding the corresponding nodes to `ignored_scopes`."
            )

    def _build_insertion_commands_list_for_quantizer_setup(
        self,
        quantizer_setup: PTQuantizerSetup,
        target_model: NNCFNetwork,
        minmax_values_for_range_init: Dict[QuantizationPointId, MinMaxTensorStatistic],
    ) -> Tuple[List[PTInsertionCommand], Dict[QuantizationPointId, QuantizerId]]:
        insertion_commands = []
        qp_id_vs_quant_module_id_dict: Dict[QuantizationPointId, QuantizerId] = {}
        target_model_graph = target_model.nncf.get_original_graph()
        non_unified_scales_quantization_point_ids = set(quantizer_setup.quantization_points.keys())
        already_weight_quantized_shared_layers: Dict[str, QuantizerId] = {}

        for unified_scales_group in quantizer_setup.unified_scale_groups.values():
            for us_qp_id in unified_scales_group:
                non_unified_scales_quantization_point_ids.discard(us_qp_id)

            (
                filtered_unified_scales_group,
                shared_weight_quantized_layers_in_group,
            ) = self._remove_shared_layer_weight_quantization_point_duplicates(
                unified_scales_group, quantizer_setup, target_model_graph
            )

            quant_module_id, commands = self._build_commands_for_single_unified_scale_group(
                target_model, quantizer_setup, filtered_unified_scales_group, minmax_values_for_range_init
            )

            for layer_name in shared_weight_quantized_layers_in_group:
                if layer_name in already_weight_quantized_shared_layers:
                    raise nncf.InternalError(
                        "Attempted to assign a unified-scale quantizer to a shared layer node that has "
                        "already had its weights quantized by another unified-scale quantizer!"
                    )
                already_weight_quantized_shared_layers[layer_name] = quant_module_id

            for us_qp_id in unified_scales_group:
                qp_id_vs_quant_module_id_dict[us_qp_id] = quant_module_id
            insertion_commands += commands

        for qp_id in non_unified_scales_quantization_point_ids:
            qp = quantizer_setup.quantization_points[qp_id]
            nncf_node = target_model_graph.get_node_by_name(qp.target_point.target_node_name)
            if qp.is_weight_quantization_point() and nncf_node.is_shared():
                layer_name = nncf_node.layer_name
                if layer_name in already_weight_quantized_shared_layers:
                    nncf_logger.debug_once(
                        f"Filtering a regular weight quantization point {qp_id} - "
                        f"already quantized as a shared layer {nncf_node.layer_name}",
                    )
                    qp_id_vs_quant_module_id_dict[qp_id] = already_weight_quantized_shared_layers[layer_name]
                    continue

            qspec = quantizer_setup.quantization_points[qp_id].qspec
            tp = quantizer_setup.quantization_points[qp_id].target_point

            range_init_minmax_values = None
            if minmax_values_for_range_init:
                minmax_stat = minmax_values_for_range_init.get(qp_id)
                if minmax_stat is not None:
                    range_init_minmax_values = (minmax_stat.min_values, minmax_stat.max_values)

            quantizer_module_id, commands = self._quantize_at_points_by_single_module(
                target_model,
                [
                    tp,
                ],
                qspec,
                range_init_minmax_values,
            )

            if (
                qp.is_weight_quantization_point()
                and nncf_node.is_shared()
                and nncf_node.layer_name not in already_weight_quantized_shared_layers
            ):
                already_weight_quantized_shared_layers[nncf_node.layer_name] = quantizer_module_id

            qp_id_vs_quant_module_id_dict[qp_id] = quantizer_module_id
            insertion_commands += commands

        adjust_padding_args = self._collect_adjust_padding_args(
            non_unified_scales_quantization_point_ids,
            qp_id_vs_quant_module_id_dict,
            quantizer_setup,
            target_model_graph,
        )

        commands = self._add_adjust_padding_ops(adjust_padding_args)
        if commands:
            insertion_commands += commands

        return insertion_commands, qp_id_vs_quant_module_id_dict

    def _remove_shared_layer_weight_quantization_point_duplicates(
        self,
        unified_scales_group: Set[QuantizationPointId],
        quantizer_setup: PTQuantizerSetup,
        target_model_graph: NNCFGraph,
    ) -> Tuple[Set[QuantizationPointId], Set[str]]:
        observed_shared_layer_names = set()
        retval = set()
        for us_qp_id in unified_scales_group:
            qp = quantizer_setup.quantization_points[us_qp_id]
            if qp.is_weight_quantization_point():
                nncf_node = target_model_graph.get_node_by_name(qp.target_point.target_node_name)
                if nncf_node.is_shared():
                    if nncf_node.layer_name not in observed_shared_layer_names:
                        observed_shared_layer_names.add(nncf_node.layer_name)
                    else:
                        nncf_logger.debug_once(
                            f"Filtering a unified-scale weight quantization point {us_qp_id} "
                            f"- already quantized as a shared layer {nncf_node.layer_name}",
                        )
                        continue
            retval.add(us_qp_id)
        return retval, observed_shared_layer_names

    def _collect_adjust_padding_args(
        self,
        non_unified_scales_quantization_point_ids: Set[QuantizationPointId],
        qp_id_vs_quant_module_id_dict: Dict[QuantizationPointId, QuantizerId],
        quantizer_setup: PTQuantizerSetup,
        target_model_graph: NNCFGraph,
    ) -> List[AdjustPaddingArgs]:
        def weight_qp_filter_fn(qp_id_):
            qp_ = quantizer_setup.quantization_points[qp_id_]
            return qp_.is_weight_quantization_point()

        weight_qps = list(filter(weight_qp_filter_fn, non_unified_scales_quantization_point_ids))
        adjust_padding_args = []
        adjust_padding_operation_set = set()
        if self.hw_config is not None:
            adjust_padding_operation_set = self.hw_config.get_operations_with_adjusted_paddings()
        for wqp_id in weight_qps:
            wqp = quantizer_setup.quantization_points[wqp_id]
            tp = quantizer_setup.quantization_points[wqp_id].target_point
            target_node = target_model_graph.get_node_by_name(tp.target_node_name)

            op_type = target_node.metatype
            is_adjust_padding_applicable = op_type in adjust_padding_operation_set
            if self._should_setup_adjust_pad_ops and is_adjust_padding_applicable:
                gid = quantizer_setup.get_shared_inputs_group_id(wqp_id)
                shared_input_group = quantizer_setup.shared_input_operation_set_groups[gid]

                def is_qp_quantizing_same_op_as_wqp(qp_id_):
                    qp_ = quantizer_setup.quantization_points[qp_id_]
                    node_matched = target_node.node_name in qp_.directly_quantized_operator_node_names
                    return qp_.is_activation_quantization_point() and node_matched

                for qp_id in filter(is_qp_quantizing_same_op_as_wqp, shared_input_group):
                    quantizer_module_id = qp_id_vs_quant_module_id_dict[qp_id]
                    activation_quantizer = self._non_weight_quantizers[quantizer_module_id].quantizer_module_ref
                    args = self._get_adjust_padding_args(
                        target_model_graph,
                        wqp,
                        activation_quantizer,
                        list(quantizer_setup.quantization_points.values()),
                    )
                    if args:
                        adjust_padding_args.extend(args)
        return adjust_padding_args

    def _build_commands_for_single_unified_scale_group(
        self,
        target_model: NNCFNetwork,
        quantizer_setup: PTQuantizerSetup,
        unified_scales_group: Set[QuantizationPointId],
        minmax_values_for_range_init: Dict[QuantizationPointId, MinMaxTensorStatistic],
    ) -> Tuple[QuantizerId, List[PTInsertionCommand]]:
        qp_ids_list_for_current_group = list(unified_scales_group)

        # The primary insertion point (to be associated with the actual quantizer module, not just hooks to it)
        # will be determined based on the string representation of said insertion point, to avoid random selection.
        # Weight insertion points are given priority.
        weight_qp_ids = [
            qp_id
            for qp_id in qp_ids_list_for_current_group
            if quantizer_setup.quantization_points[qp_id].is_weight_quantization_point()
        ]
        act_qp_ids = [
            qp_id
            for qp_id in qp_ids_list_for_current_group
            if quantizer_setup.quantization_points[qp_id].is_activation_quantization_point()
        ]

        def ip_str_repr_key_lambda(x):
            return str(quantizer_setup.quantization_points[x].target_point.target_node_name)

        sorted_wqp_ids = sorted(weight_qp_ids, key=ip_str_repr_key_lambda)
        sorted_aqp_ids = sorted(act_qp_ids, key=ip_str_repr_key_lambda)
        sorted_qp_ids = sorted_wqp_ids + sorted_aqp_ids

        primary_qp_id = sorted_qp_ids[0]
        linked_qp_ids = sorted_qp_ids[1:]
        qspec = quantizer_setup.quantization_points[primary_qp_id].qspec
        linked_qspecs = [quantizer_setup.quantization_points[qp_id].qspec for qp_id in linked_qp_ids]
        for linked_qspec in linked_qspecs:
            if not qspec == linked_qspec:
                raise nncf.InternalError("The qspecs for unified scale quantization points should be identical!")

        range_init_minmax_values = None
        if minmax_values_for_range_init:
            # Hopefully this will suffice.
            # TODO: gather unified statistic by linking stat collectors_and_modules_to_init instead
            min_values = None
            max_values = None
            for qp_id in sorted_qp_ids:
                minmax_stat = minmax_values_for_range_init.get(qp_id)
                if minmax_stat is None:
                    continue

                if min_values is None:
                    min_values = minmax_stat.min_values.data
                else:
                    min_values = torch.min(min_values, minmax_stat.min_values.data)

                if max_values is None:
                    max_values = minmax_stat.max_values.data
                else:
                    max_values = torch.max(max_values, minmax_stat.max_values.data)
            if min_values is not None and max_values is not None:
                range_init_minmax_values = min_values, max_values

        target_points = [quantizer_setup.quantization_points[qp_id].target_point for qp_id in sorted_qp_ids]
        quantizer_module_id, commands = self._quantize_at_points_by_single_module(
            target_model, target_points, qspec, range_init_minmax_values
        )
        return quantizer_module_id, commands

    def _select_final_qconfig(self, quantizer_config_list: List[QuantizerConfig]) -> QuantizerConfig:
        # Quantizer config list entries should arrive in the same order as they are listed
        # in the HW config, where they are sorted by descending order of priority
        return quantizer_config_list[0]

    def _quantize_at_points_by_single_module(
        self,
        target_model: NNCFNetwork,
        insertion_points: List[PTTargetPoint],
        qspec: PTQuantizerSpec,
        range_init_minmax_values: Tuple[torch.Tensor, torch.Tensor] = None,
    ) -> Tuple[QuantizerId, List[PTInsertionCommand]]:
        """
        Will generate insertion commands for quantization at possibly multiple points
        in the network using one and the same trainable quantizer module. The trainable
        quantizer module will be saved either inside the weightable module which weights
        it quantizes (for single-point weight quantization), or in a NNCFNetwork wrapper
        module (i.e. in a storage external to the original module).
        :param: target_model - the model to be quantized.
        :param: insertion_points - a list of target points for quantization using one
        quantizer module
        :param: qconfig - the QuantizerConfig for the resulting quantizer module
        :param: range_init_minmax_values - a pair of minimum and maximum values of input statistics
        for initializing the quantizer's trainable parameters
        :return: A tuple with the identifier of the new quantizer module and a list of
        insertion commands registering this module for quantization at spots described by
        insertion_points.
        """

        target_model_graph = target_model.nncf.get_original_graph()
        if not insertion_points:
            raise nncf.InternalError("No insertion points to put quantizers into!")

        def is_weights(ip: PTTargetPoint) -> bool:
            return ip.target_type is TargetType.OPERATION_WITH_WEIGHTS

        primary_ip = insertion_points[0]

        quantizer = self.__create_quantize_module(qspec).to(self._device_for_callable_obj_creation)
        if range_init_minmax_values is not None:
            # Need to cast to the model's current dtype since the statistics could have been gathered in an
            # AMP autocast model (and therefore be FP16 since AMP autocast switches precision of activations
            # at forward pass time)
            own_type = get_model_dtype(target_model)
            min_values = range_init_minmax_values[0].data.type(own_type)
            max_values = range_init_minmax_values[1].data.type(own_type)

            quantizer.apply_minmax_init(min_values=min_values, max_values=max_values, log_module_name=str(primary_ip))

        qids: List[QuantizerId] = []
        for ip in insertion_points:
            if is_weights(ip):
                qids.append(WeightQuantizerId(ip.target_node_name))
            else:
                qids.append(NonWeightQuantizerId(ip.target_node_name, ip.input_port_id))

        serialized_insertions_list = [str(x) for x in qids]
        external_quantizer_storage_key = ";".join(serialized_insertions_list)
        if len(insertion_points) > 1:
            linked_quantizers_str = "\n".join(serialized_insertions_list)
            nncf_logger.info_once(f"Scales will be unified for quantizer group:\n{linked_quantizers_str}\n")

        if is_weights(primary_ip):
            primary_qid = WeightQuantizerId(primary_ip.target_node_name)
            self._weight_quantizers[primary_qid] = WeightQuantizerInfo(
                quantizer, target_model.nncf.get_containing_module(primary_ip.target_node_name), insertion_points
            )
            module_node = target_model_graph.get_node_by_name(primary_ip.target_node_name)
            layer_attributes = module_node.layer_attributes
            input_shape = layer_attributes.get_weight_shape()
            self._quantizers_input_shapes[primary_qid] = tuple(input_shape)
        else:
            primary_qid = NonWeightQuantizerId(primary_ip.target_node_name, primary_ip.input_port_id)
            self._non_weight_quantizers[primary_qid] = NonWeightQuantizerInfo(quantizer, insertion_points)
            input_shape = target_model_graph.get_input_shape_for_insertion_point(insertion_points[0])
            self._quantizers_input_shapes[primary_qid] = input_shape

        if not (is_weights(primary_ip) and len(insertion_points) == 1):
            assert external_quantizer_storage_key not in target_model.nncf.get_compression_modules_by_type(
                ExtraCompressionModuleType.EXTERNAL_QUANTIZER
            )

            target_model.nncf.add_compression_module(
                external_quantizer_storage_key, quantizer, ExtraCompressionModuleType.EXTERNAL_QUANTIZER
            )

        insertion_commands = []
        for curr_insertion_point in insertion_points:
            if curr_insertion_point in self._processed_insertion_points:
                raise nncf.InternalError("Insertion point {} already quantized!".format(str(curr_insertion_point)))
            self._processed_insertion_points.add(curr_insertion_point)

            if is_weights(curr_insertion_point):
                if len(insertion_points) == 1:
                    # For backward compatibility, if only one weight is quantized by a single quantizer,
                    # insert UpdateWeight ops with a genuine quantizer module
                    callable_obj = quantizer
                else:
                    # Otherwise use external quantizer module storage since the quantization points will have to
                    # share the single module and this would be impossible for multiple weight quantizer sharing if
                    # the corresponding UpdateWeights operations contained real modules (these would simply get copied
                    # by PyTorch internals)
                    callable_obj = ExternalQuantizerCallHook(external_quantizer_storage_key, self._debug_interface)
            else:
                # Hooks will be identical for each affected op_address in the linked scenario
                # - will call one and the same quantizer
                callable_obj = ExternalQuantizerCallHook(external_quantizer_storage_key, self._debug_interface)

            nncf_logger.debug_once(
                f"Performing "
                f"{'signed' if quantizer.signed else 'unsigned'} "
                f"{'logarithm_scale' if quantizer.is_using_log_scale_storage else ''} "
                f"{'weight' if is_weights(curr_insertion_point) else 'activation'} "
                f"quantization for: {str(curr_insertion_point)}",
            )

            insertion_commands.append(
                PTInsertionCommand(curr_insertion_point, callable_obj, TransformationPriority.QUANTIZATION_PRIORITY)
            )
        return primary_qid, insertion_commands

    def _are_frozen_layers_allowed(self) -> Tuple[bool, str]:
        message_template = Template("Frozen layers are$denial allowed for $algo_prefix quantization")
        bits = set()
        bits.update({wq.quantizer_module_ref.num_bits for wq in self._weight_quantizers.values()})
        bits.update({nwq.quantizer_module_ref.num_bits for nwq in self._non_weight_quantizers.values()})

        if self._precision_init_params or len(bits) > 1:
            return False, message_template.substitute(denial=" not", algo_prefix="mixed precision")

        if len(bits) == 1:
            bitwidth = bits.pop()
            algo_prefix = f"INT{bitwidth}"
            if bitwidth == 8:
                return True, message_template.substitute(denial="", algo_prefix=algo_prefix)
            return False, message_template.substitute(denial=" not", algo_prefix=algo_prefix)
        return True, message_template.substitute(denial="", algo_prefix="empty")

    def _get_compression_lr_multiplier(self) -> Optional[float]:
        return self.config.get_redefinable_global_param_value_for_algo("compression_lr_multiplier", self.name)

    def initialize(self, model: NNCFNetwork) -> None:
        if is_main_process() and self.should_init:
            bn_adapt_params = self._parse_bn_adapt_params()
            if bn_adapt_params is not None:
                bn_adaptation = BatchnormAdaptationAlgorithm(
                    **extract_bn_adaptation_init_params(self.config, "quantization")
                )
                bn_adaptation.run(model)


@api()
class QuantizationController(QuantizationControllerBase):
    """
    Controller for the quantization algorithm in PT.
    """

    def __init__(
        self,
        target_model: NNCFNetwork,
        config: NNCFConfig,
        debug_interface: "QuantizationDebugInterface",
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        non_weight_quantizers: Dict[NonWeightQuantizerId, NonWeightQuantizerInfo],
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
        quantizers_input_shapes: Dict[QuantizerId, Tuple[int]],
        build_time_metric_info: QuantizationShareBuildTimeInfo = None,
        build_time_range_init_params: PTRangeInitParams = None,
    ):
        super().__init__(target_model)
        self._loss = ZeroCompressionLoss(get_model_device(target_model))
        self._scheduler = BaseCompressionScheduler()
        self.debug_interface = debug_interface
        self.config = config
        algo_config = self._get_algo_config()
        self._build_time_range_init_params = build_time_range_init_params

        self.weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo] = weight_quantizers
        self.non_weight_quantizers: Dict[NonWeightQuantizerId, NonWeightQuantizerInfo] = non_weight_quantizers
        self.all_quantizations: Dict[QuantizerId, BaseQuantizer] = OrderedDict()
        self.all_quantizations.update({k: v.quantizer_module_ref for k, v in self.weight_quantizers.items()})
        self.all_quantizations.update({k: v.quantizer_module_ref for k, v in self.non_weight_quantizers.items()})
        self._quantizers_input_shapes = quantizers_input_shapes
        self._distributed = False
        self._groups_of_adjacent_quantizers = groups_of_adjacent_quantizers
        self._bn_adaptation = None
        self._build_time_metric_info = build_time_metric_info
        self._target_device = self.config.get("target_device", "ANY")

        should_export_to_onnx_qdq = algo_config.get(
            "export_to_onnx_standard_ops", QUANTIZATION_EXPORT_TO_ONNX_STANDARD_OPS
        )
        if should_export_to_onnx_qdq:
            warning_deprecated(
                "The config option `export_to_onnx_standard_ops` is deprecated and will be removed "
                "in a future version. Please use the `nncf.strip(quantized_model)` method before export to ONNX "
                "to get model with QuantizeLinear-DequantizeLinear node pairs."
            )
            export_mode = QuantizerExportMode.ONNX_QUANTIZE_DEQUANTIZE_PAIRS
        else:
            export_mode = QuantizerExportMode.FAKE_QUANTIZE

        for quantizer in self.all_quantizations.values():
            quantizer.set_export_mode(export_mode)

        params = algo_config.get("params", None)
        self.is_staged_scheduler = bool(params)

        # Staged scheduler must be created after initialized to prevent extra logic with disabled quantizations
        if self.is_staged_scheduler:
            scheduler_cls = QUANTIZATION_SCHEDULERS.get("staged")
            self._scheduler = scheduler_cls(self, params)

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    @property
    def groups_of_adjacent_quantizers(self) -> GroupsOfAdjacentQuantizers:
        return self._groups_of_adjacent_quantizers

    def prepare_for_export(self):
        for quantizer_id, quantizer in self.all_quantizations.items():
            if not quantizer.is_enabled_quantization():
                nncf_logger.debug(f"Disabled quantization on export to ONNX: {quantizer_id}")

    def distributed(self):
        self._distributed = True
        self._broadcast_initialized_params_for_each_quantizer()

    def _get_algo_config(self) -> Dict:
        return extract_algo_specific_config(self.config, "quantization")

    def _broadcast_initialized_params_for_each_quantizer(self):
        # NOTE: Order of quantization modules must be the same on GPUs to correctly broadcast num_bits
        sorted_quantizers = OrderedDict(sorted(self.all_quantizations.items(), key=lambda x: str(x[0])))
        for quantizer in sorted_quantizers.values():
            quantizer.broadcast_initialized_params()

    def _do_runtime_range_init(self, range_init_params: PTRangeInitParams):
        modules_to_init = OrderedDict()
        for wq_id, wq_info in self.weight_quantizers.items():
            group = QuantizerGroup.WEIGHTS
            init_config = range_init_params.get_init_config_for_scope_and_group(wq_id, group)
            is_weights = True
            modules_to_init[str(wq_id)] = (
                wq_info.quantizer_module_ref,
                init_config,
                is_weights,
                self._quantizers_input_shapes[wq_id],
            )

        for aq_id, aq_info in self.non_weight_quantizers.items():
            group = QuantizerGroup.ACTIVATIONS
            init_config = range_init_params.get_init_config_for_scope_and_group(aq_id, group)
            is_weights = False
            modules_to_init[str(aq_id)] = (
                aq_info.quantizer_module_ref,
                init_config,
                is_weights,
                self._quantizers_input_shapes[aq_id],
            )

        # NOTE: Order of modules must be the same to correctly broadcast parameters (e.g. input_low
        # and input_range)
        modules_to_init = OrderedDict(sorted(modules_to_init.items()))
        self.modules_to_range_init = modules_to_init
        runner = DataLoaderRangeInitializeRunner(self._model, modules_to_init, range_init_params.device)

        quantizers = [module for module, config, is_weights, input_shape in modules_to_init.values()]
        quantizers_switcher = QuantizersSwitcher(quantizers)
        # bypass quantization to collect statistics from floating point model
        quantizers_switcher.disable_quantizers()
        with training_mode_switcher(self._model, is_training=False):
            # Statistics should be collected in eval mode because the model in train mode may behave differently
            runner.run(range_init_params.init_range_data_loader, range_init_params.get_max_num_init_steps())
        quantizers_switcher.enable_quantizers()

        self._model.nncf.rebuild_graph()

    def compression_stage(self) -> CompressionStage:
        if self.is_staged_scheduler:
            return self.scheduler.compression_stage()
        return CompressionStage.FULLY_COMPRESSED

    def init_precision(
        self,
        precision_init_type: str,
        precision_init_params: BasePrecisionInitParams,
        precision_constraints: HardwareQuantizationConstraints,
    ) -> SingleConfigQuantizerSetup:
        """
        Precision initialization happens based on an measure of layer sensitivity to perturbations. The measure is
        calculated by average Hessian trace estimation for each layer using Hutchinson algorithm.
        """
        init_impl = PrecisionInitializerFactory.create(precision_init_type)
        initializer = init_impl(self, precision_init_params, precision_constraints)
        nncf_logger.info("Initializing quantizer precisions...")
        return initializer.apply_init()

    def init_range(self, range_init_params: PTRangeInitParams = None):
        """
        Tracks input statistics for quantizers in the model and sets ranges of the quantizers to correspond to
        minimum and maximum input tensor levels observed.
        :param range_init_params: specifies parameters for this range initialization call; if None, the parameters
        that were used during compressed model creation will be used.
        """
        if range_init_params is None:
            if self._build_time_range_init_params is None:
                nncf_logger.error(
                    "Requested a quantization controller to do range initialization without "
                    "`range_init_params` function parameter supplied, but the build time range "
                    "initialization was not supplied with params as well. "
                    "Range initialization will not be done."
                )
                return
            range_init_params = self._build_time_range_init_params

        self._do_runtime_range_init(range_init_params)

        if self._distributed:
            self._broadcast_initialized_params_for_each_quantizer()

    def enable_activation_quantization(self):
        for m in self.non_weight_quantizers.values():
            m.quantizer_module_ref.enable_quantization()

    def enable_weight_quantization(self):
        for m in self.weight_quantizers.values():
            m.quantizer_module_ref.enable_quantization()

    def disable_activation_quantization(self):
        for m in self.non_weight_quantizers.values():
            m.quantizer_module_ref.disable_quantization()

    def disable_weight_quantization(self):
        for m in self.weight_quantizers.values():
            m.quantizer_module_ref.disable_quantization()

    def statistics(self, quickly_collected_only=False) -> NNCFStatistics:
        if not quickly_collected_only and is_debug():
            stats = MemoryConsumptionStatisticsCollector(
                self.model, self.weight_quantizers, self.non_weight_quantizers
            ).collect()
            nncf_logger.debug(stats.to_str())

            stats = ShareEdgesQuantizedDataPathStatisticsCollector(self.model, self, self._target_device).collect()
            nncf_logger.debug(stats.to_str())

        collector = PTQuantizationStatisticsCollector(
            self.weight_quantizers, self.non_weight_quantizers, self._build_time_metric_info
        )
        stats = collector.collect()

        nncf_stats = NNCFStatistics()
        nncf_stats.register("quantization", stats)
        return nncf_stats

    def strip_model(self, model: NNCFNetwork, do_copy: bool = False) -> NNCFNetwork:
        if do_copy:
            model = copy_model(model)
        model = strip_quantized_model(model)
        return model


class ExperimentalQuantizationBuilder(QuantizationBuilder):
    def __init__(
        self,
        quantizer_setup: MultiConfigQuantizerSetup,
        initial_quantizer_setup: SingleConfigQuantizerSetup,
        tensor_stats_for_all_setup_variations: Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]],
        hw_config: HWConfig = None,
    ):
        should_init = bool(tensor_stats_for_all_setup_variations)
        super().__init__(NNCFConfig(), should_init=should_init)
        self._initial_quantizer_setup = initial_quantizer_setup
        self._quantizer_setup = quantizer_setup
        self._tensor_stats = tensor_stats_for_all_setup_variations
        self._should_setup_adjust_pad_ops = False
        self.hw_config = hw_config

    def _handle_frozen_layers(self, target_model: NNCFNetwork):
        pass

    def _get_single_config_quantizer_setup(self, target_model) -> SingleConfigQuantizerSetup:
        return self._initial_quantizer_setup

    def _get_statistics_for_final_range_init(
        self, target_model: NNCFNetwork, quantizer_setup: QuantizerSetupBase, range_init_params: PTRangeInitParams
    ) -> Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]]:
        return self._tensor_stats

    def _build_controller(self, model: NNCFNetwork) -> "ExperimentalQuantizationController":
        groups_of_adjacent_quantizers = GroupsOfAdjacentQuantizers()
        all_quantizations: Dict[QuantizerId, BaseQuantizer] = {}
        all_quantizations.update({k: v.quantizer_module_ref for k, v in self._weight_quantizers.items()})
        all_quantizations.update({k: v.quantizer_module_ref for k, v in self._non_weight_quantizers.items()})

        groups_of_adjacent_quantizers.parse_from_quantizer_setup(
            all_quantizations, self._pt_quantizer_setup, self._setup_to_module_id_translation_dict
        )

        build_time_metric_infos = QuantizationShareBuildTimeInfo(
            len(self._non_weight_quantizers), len(self._weight_quantizers)
        )

        return ExperimentalQuantizationController(
            model,
            self._weight_quantizers,
            self._non_weight_quantizers,
            groups_of_adjacent_quantizers,
            self._quantizers_input_shapes,
            self._quantizer_setup,
            self._initial_quantizer_setup,
            self._setup_to_module_id_translation_dict,
            self._tensor_stats,
            build_time_metric_infos,
            self._should_setup_adjust_pad_ops,
            self.hw_config,
        )

    def initialize(self, model: NNCFNetwork) -> None:
        pass

    def _get_algo_specific_config_section(self) -> Dict:
        return {}

    def _parse_range_init_params(self) -> Optional[PTRangeInitParams]:
        return None

    def _get_compression_lr_multiplier(self) -> Optional[float]:
        return None


class ExperimentalQuantizationController(QuantizationController):
    def __init__(
        self,
        target_model: NNCFNetwork,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        non_weight_quantizers: Dict[NonWeightQuantizerId, NonWeightQuantizerInfo],
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
        quantizers_input_shapes: Dict[QuantizerId, Tuple[int]],
        quantizer_setup: MultiConfigQuantizerSetup,
        initial_quantizer_setup: SingleConfigQuantizerSetup,
        setup_to_module_id_translation_dict: Dict[QuantizationPointId, QuantizerId],
        tensor_stats: Dict[PTTargetPoint, Dict[ReductionAxes, TensorStatistic]],
        build_time_metric_info: QuantizationShareBuildTimeInfo,
        should_setup_adjust_pad_ops=False,
        hw_config: HWConfig = None,
    ):
        super().__init__(
            target_model,
            NNCFConfig(),
            debug_interface=None,
            weight_quantizers=weight_quantizers,
            non_weight_quantizers=non_weight_quantizers,
            groups_of_adjacent_quantizers=groups_of_adjacent_quantizers,
            quantizers_input_shapes=quantizers_input_shapes,
            build_time_metric_info=build_time_metric_info,
        )
        self._target_model_ref = target_model
        self._should_setup_adjust_pad_ops = should_setup_adjust_pad_ops
        self._quantizer_setup = quantizer_setup
        self._initial_quantizer_setup = initial_quantizer_setup
        self._tensor_stats = tensor_stats
        self.setup_to_module_id_translation_dict = setup_to_module_id_translation_dict
        self.module_id_to_qp_id_translation_dict: Dict[QuantizerId, Set[QuantizationPointId]] = {}
        for qp_id, qid in self.setup_to_module_id_translation_dict.items():
            if qid in self.module_id_to_qp_id_translation_dict:
                self.module_id_to_qp_id_translation_dict[qid].add(qp_id)
            else:
                self.module_id_to_qp_id_translation_dict[qid] = {qp_id}
        self.hw_config = hw_config

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    def get_quantizer_setup_for_current_state(self) -> SingleConfigQuantizerSetup:
        qpid_vs_selected_qconfig = {}
        for qp_id in self._initial_quantizer_setup.quantization_points:
            quant_module_id = self.setup_to_module_id_translation_dict[qp_id]
            quant_module = self.all_quantizations[quant_module_id]
            qconfig = quant_module.get_quantizer_config()
            qpid_vs_selected_qconfig[qp_id] = qconfig
        return self._quantizer_setup.select_qconfigs(qpid_vs_selected_qconfig, strict=False)

    def is_new_setup_requires_regeneration(self, quantizer_setup: SingleConfigQuantizerSetup) -> bool:
        current_setup = self.get_quantizer_setup_for_current_state()
        if Counter(current_setup.quantization_points.keys()) != Counter(quantizer_setup.quantization_points.keys()):
            raise ValueError("The new setup is inconsistent with the original parameter space!")
        for qp_id, qp in quantizer_setup.quantization_points.items():
            current_qconfig = current_setup.quantization_points[qp_id].qconfig
            new_qconfig = quantizer_setup.quantization_points[qp_id].qconfig
            new_padding_adjust_applicable = CalculatePaddingAdjustment.is_config_applicable(new_qconfig)
            current_padding_adjust_applicable = CalculatePaddingAdjustment.is_config_applicable(current_qconfig)
            need_padding_regeneration = (
                self._should_setup_adjust_pad_ops
                and qp.is_activation_quantization_point()
                and new_padding_adjust_applicable != current_padding_adjust_applicable
            )
            if (
                current_qconfig.per_channel != new_qconfig.per_channel
                or (
                    new_qconfig.signedness_to_force is not None
                    and current_qconfig.signedness_to_force != new_qconfig.signedness_to_force
                )
                or current_qconfig.mode != new_qconfig.mode
                or need_padding_regeneration
            ):
                return True
        return False

    def apply_new_quantizer_setup(
        self, quantizer_setup: SingleConfigQuantizerSetup
    ) -> Tuple["ExperimentalQuantizationController", NNCFNetwork]:
        if not self.is_new_setup_requires_regeneration(quantizer_setup):
            for qp_id, qp in quantizer_setup.quantization_points.items():
                quant_module_id = self.setup_to_module_id_translation_dict[qp_id]
                quant_module = self.all_quantizations[quant_module_id]
                quant_module.num_bits = qp.qconfig.num_bits
            return self, self._target_model_ref
        new_model = self._target_model_ref.nncf.get_clean_shallow_copy()
        new_builder = ExperimentalQuantizationBuilder(
            self._quantizer_setup,
            initial_quantizer_setup=quantizer_setup,
            tensor_stats_for_all_setup_variations=self._tensor_stats,
            hw_config=self.hw_config,
        )
        new_builder.apply_to(new_model)
        new_ctrl = new_builder.build_controller(new_model)
        return new_ctrl, new_model

    def _get_algo_config(self) -> Dict:
        return {}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict
from typing import Any, Dict, List, Optional

import openvino.runtime as ov

from nncf.common.logging import nncf_logger
from nncf.scopes import IgnoredScope


def exclude_empty_fields(value: Dict[str, Any]) -> Dict[str, Any]:
    """
    Remove keys where value is empty and key is `validate`

    :param value: IgnoredScope instance, based on the quantization parameters
    """
    value.pop("validate")
    keys = list(value.keys())
    for key in keys:
        if not value[key]:
            value.pop(key)
    return value


def dump_parameters(
    model: ov.Model, parameters: Dict, algo_name: Optional[str] = "quantization", path: Optional[List] = None
) -> None:
    """
    Dumps the given parameters into Model's meta section.

    :param model: ov.Model instance.
    :param algo_name: Name of the algorithm, to which the parameters refer.
    :param parameters: Incoming dictionary with parameters to save.
    :param path: Optional list of the paths.
    """
    try:
        path = path if path else []
        for key, value in parameters.items():
            # Special condition for composed fields like IgnoredScope
            if isinstance(value, IgnoredScope):
                value = exclude_empty_fields(asdict(value))
                if bool(value):
                    dump_parameters(model, value, algo_name, [key])
                    continue
                else:
                    # The default value in case empty ignored_scope parameter passed
                    value = []

            rt_path = ["nncf", algo_name] + path + [key]
            model.set_rt_info(str(value), rt_path)
    except RuntimeError as e:
        nncf_logger.debug(f"Unable to dump optimization parameters due to error: {e}")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

import torch
from torch import nn

from nncf import NNCFConfig
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.utils import is_prunable_depthwise_conv
from nncf.common.utils.helpers import create_table
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.schemata.defaults import PRUNE_BATCH_NORMS
from nncf.config.schemata.defaults import PRUNE_DOWNSAMPLE_CONVS
from nncf.config.schemata.defaults import PRUNE_FIRST_CONV
from nncf.torch.algo_selector import ZeroCompressionLoss
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TransformationPriority
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.module_operations import UpdateWeightAndBias
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.pruning.operations import PT_PRUNING_OPERATOR_METATYPES
from nncf.torch.pruning.structs import PrunedModuleInfo
from nncf.torch.pruning.tensor_processor import PTNNCFPruningTensorProcessor
from nncf.torch.pruning.utils import init_output_masks_in_graph
from nncf.torch.utils import get_model_device


class BasePruningAlgoBuilder(PTCompressionAlgorithmBuilder):
    def __init__(self, config, should_init: bool = True):
        super().__init__(config, should_init)
        params = self._algo_config.get("params", {})
        self._set_default_params_for_ranking_type(params)
        self._params = params

        self.prune_first = params.get("prune_first_conv", PRUNE_FIRST_CONV)
        self.prune_batch_norms = params.get("prune_batch_norms", PRUNE_BATCH_NORMS)
        self.prune_downsample_convs = params.get("prune_downsample_convs", PRUNE_DOWNSAMPLE_CONVS)

        self._prunable_types = self.get_op_types_of_pruned_modules()

        from nncf.common.pruning.node_selector import PruningNodeSelector

        self.pruning_node_selector = PruningNodeSelector(
            PT_PRUNING_OPERATOR_METATYPES,
            self._prunable_types,
            self.get_types_of_grouping_ops(),
            self.ignored_scopes,
            self.target_scopes,
            self.prune_first,
            self.prune_downsample_convs,
        )

        self.pruned_module_groups_info = []
        self._pruned_norms_operators = []

    @staticmethod
    def _set_default_params_for_ranking_type(params: Dict) -> None:
        """
        Setting default parameter values of pruning algorithm depends on the ranking type:
        for learned_ranking `all_weights` must be True (in case of False was set by the user, an Exception will be
        raised), `prune_first_conv`, `prune_downsample_convs` are recommended to be True (this
        params will be set to True by default (and remain unchanged if the user sets some value).
        :param params: dict with parameters of the algorithm from config
        """
        learned_ranking = "interlayer_ranking_type" in params and params["interlayer_ranking_type"] == "learned_ranking"
        if not learned_ranking:
            return
        nncf_logger.info(
            "LeGR pruning sets `prune_first_conv`, `prune_downsample_convs`, "
            "`all_weights` to True by default. Adjusting this is not recommended."
        )
        params.setdefault("prune_first_conv", True)
        params.setdefault("prune_downsample_convs", True)
        if params.get("all_weights") is False:
            raise Exception(
                "For LeGR pruning the `all_weights` config parameter be set to `true`."
                "Adjust the config accordingly if you want to proceed."
            )
        params.setdefault("all_weights", True)

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        layout = PTTransformationLayout()
        commands = self._prune_weights(target_model)
        for command in commands:
            layout.register(command)
        return layout

    def _prune_weights(self, target_model: NNCFNetwork):
        target_model_graph = target_model.nncf.get_original_graph()
        groups_of_nodes_to_prune = self.pruning_node_selector.create_pruning_groups(target_model_graph)

        device = get_model_device(target_model)
        insertion_commands = []
        self.pruned_module_groups_info = Clusterization[PrunedModuleInfo](lambda x: x.node_name)

        for i, group in enumerate(groups_of_nodes_to_prune.get_all_clusters()):
            group_minfos = []
            for node in group.elements:
                node_name = node.node_name
                module = target_model.nncf.get_containing_module(node_name)
                module_scope = target_model_graph.get_scope_by_node_name(node_name)
                # Check that we need to prune weights in this op
                assert self._is_pruned_module(module)

                nncf_logger.debug(f"Will prune the weights for the operation: {node_name}")
                pruning_block = self.create_weight_pruning_operation(module, node_name)
                # Hook for weights and bias
                hook = UpdateWeightAndBias(pruning_block).to(device)
                insertion_commands.append(
                    PTInsertionCommand(
                        PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                        hook,
                        TransformationPriority.PRUNING_PRIORITY,
                    )
                )
                group_minfos.append(
                    PrunedModuleInfo(
                        node_name=node_name,
                        module_scope=module_scope,
                        module=module,
                        operand=pruning_block,
                        node_id=node.node_id,
                        is_depthwise=is_prunable_depthwise_conv(node),
                    )
                )

            cluster = Cluster[PrunedModuleInfo](i, group_minfos, [n.node_id for n in group.elements])
            self.pruned_module_groups_info.add_cluster(cluster)

        # Propagate masks to find norm layers to prune
        init_output_masks_in_graph(target_model_graph, self.pruned_module_groups_info.get_all_nodes())
        MaskPropagationAlgorithm(
            target_model_graph, PT_PRUNING_OPERATOR_METATYPES, PTNNCFPruningTensorProcessor
        ).mask_propagation()

        # Adding binary masks also for Batch/Group/Layer Norms to allow applying masks after propagation
        types_to_apply_mask = ["group_norm", "layer_norm"]
        if self.prune_batch_norms:
            types_to_apply_mask.append("batch_norm")

        all_norm_layers = target_model_graph.get_nodes_by_types(types_to_apply_mask)
        for node in all_norm_layers:
            if node.attributes["output_mask"] is None:
                # Skip elements that will not be pruned
                continue

            node_name = node.node_name
            module = target_model.nncf.get_containing_module(node_name)

            pruning_block = self.create_weight_pruning_operation(module, node_name)
            # Hook for weights and bias
            hook = UpdateWeightAndBias(pruning_block).to(device)
            insertion_commands.append(
                PTInsertionCommand(
                    PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                    hook,
                    TransformationPriority.PRUNING_PRIORITY,
                )
            )
            self._pruned_norms_operators.append((node, pruning_block, module))
        return insertion_commands

    def create_weight_pruning_operation(self, module, node_name):
        raise NotImplementedError

    def _is_pruned_module(self, module: nn.Module):
        """
        Return whether this module should be pruned or not.
        """
        raise NotImplementedError

    def get_op_types_of_pruned_modules(self):
        """
        Returns list of operation types that should be pruned.
        """
        raise NotImplementedError

    def get_types_of_grouping_ops(self):
        raise NotImplementedError

    def initialize(self, model: NNCFNetwork) -> None:
        pass


class BasePruningAlgoController(PTCompressionAlgorithmController):
    def __init__(
        self,
        target_model: NNCFNetwork,
        prunable_types: List[str],
        pruned_module_groups_info: Clusterization[PrunedModuleInfo],
        config: NNCFConfig,
    ):
        super().__init__(target_model)
        self._loss = ZeroCompressionLoss(get_model_device(target_model))
        self._prunable_types = prunable_types
        self.config = config
        self.pruning_config = extract_algo_specific_config(config, "filter_pruning")
        params = self.pruning_config.get("params", {})
        self.pruned_module_groups_info = pruned_module_groups_info
        self.prune_batch_norms = params.get("prune_batch_norms", PRUNE_BATCH_NORMS)
        self.prune_first = params.get("prune_first_conv", PRUNE_FIRST_CONV)
        self.prune_downsample_convs = params.get("prune_downsample_convs", PRUNE_DOWNSAMPLE_CONVS)
        self.prune_flops = False
        self.check_pruning_level(params)
        self._hooks = []

    def freeze(self):
        raise NotImplementedError

    def set_pruning_level(self, pruning_level):
        raise NotImplementedError

    def step(self, next_step):
        pass

    def check_pruning_level(self, params):
        """
        Check that set only one of pruning target params
        """
        pruning_target = params.get("pruning_target", None)
        pruning_flops_target = params.get("pruning_flops_target", None)
        if pruning_target and pruning_flops_target:
            raise ValueError("Only one parameter from 'pruning_target' and 'pruning_flops_target' can be set.")
        if pruning_flops_target:
            self.prune_flops = True

    def _clean_hooks(self):
        for h in self._hooks:
            h.remove()
        self._hooks = []

    def get_mask(self, minfo: PrunedModuleInfo) -> torch.Tensor:
        """
        Returns pruning mask for minfo.module.
        """
        raise NotImplementedError

    def pruning_level_for_mask(self, minfo: PrunedModuleInfo):
        mask = self.get_mask(minfo)
        pruning_level = 1 - mask.nonzero().size(0) / max(mask.view(-1).size(0), 1)
        return pruning_level

    def mask_shape(self, minfo: PrunedModuleInfo):
        mask = self.get_mask(minfo)
        return mask.shape

    def get_stats_for_pruned_modules(self):
        """
        Creates a table with layer pruning level statistics
        """
        header = ["Name", "Weight's shape", "Bias shape", "Layer PR"]
        rows_data = []
        for minfo in self.pruned_module_groups_info.get_all_nodes():
            rows_data.append(
                [
                    str(minfo.module_scope),
                    list(minfo.module.weight.size()),
                    list(minfo.module.bias.size()) if minfo.module.bias is not None else [],
                    self.pruning_level_for_mask(minfo),
                ]
            )

        return create_table(header, rows_data, max_col_widths=[33, 20, 6, 8])



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from enum import Enum
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, TypeVar, Union

import numpy as np

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.pruning.symbolic_mask import SymbolicMask
from nncf.common.utils.registry import Registry


def is_grouped_conv(node: NNCFNode) -> bool:
    """
    Returns `True` if a feeded node is a grouped convolution node.

    :param node: NNCFNode to check.
    :return: `True` if a feeded node a grouped convolution node.
    """
    return isinstance(node.layer_attributes, ConvolutionLayerAttributes) and node.layer_attributes.groups != 1


def is_batched_linear(node: NNCFNode, graph: NNCFGraph) -> bool:
    """
    Returns `True` if a feeded linear node output tensor has no more than two dimensions.
    A linear layer has more than two output dimensions means, that this
    linear layer multiplies several input matrices feeded by batch dimensions
    from the left/right or both inputs. Batch input dimensions are elements of [:-2] slice.

    :param node: NNCFNode to check.
    :param graph: NNCFGraph which feeded node is belonged to.
    :return: `True` if a feeded linear node output tensor has no more than two dimensions.
    """
    if not isinstance(node.layer_attributes, LinearLayerAttributes):
        return False

    edges = graph.get_output_edges(node)
    if not edges:
        return False

    return len(edges[0].tensor_shape) > 2


def get_sources_of_node(nncf_node: NNCFNode, graph: NNCFGraph, sources_types: List[str]) -> List[NNCFNode]:
    """
    Source is a node of source such that there is path from this node to `nncf_node` and on this path
    no node has one of `sources_types` type.

    :param sources_types: List of sources types.
    :param nncf_node: NNCFNode to get sources.
    :param graph: NNCF graph to work with.
    :return: List of all sources nodes.
    """
    visited = {node_id: False for node_id in graph.get_all_node_ids()}
    partial_traverse_function = partial(traverse_function, type_check_fn=lambda x: x in sources_types, visited=visited)
    nncf_nodes = [nncf_node]
    if nncf_node.node_type in sources_types:
        nncf_nodes = graph.get_previous_nodes(nncf_node)

    source_nodes = []
    for node in nncf_nodes:
        source_nodes.extend(graph.traverse_graph(node, partial_traverse_function, False))
    return source_nodes


def find_next_nodes_not_of_types(graph: NNCFGraph, nncf_node: NNCFNode, types: List[str]) -> List[NNCFNode]:
    """
    Traverse nodes in the graph from nncf node to find first nodes that aren't of type from types list.
    First nodes with some condition mean nodes:
        - for which this condition is true
        - reachable from `nncf_node` such that on the path from `nncf_node` to
          this nodes there are no other nodes with fulfilled condition

    :param graph: Graph to work with.
    :param nncf_node: NNCFNode to start search.
    :param types: List of types.
    :return: List of next nodes for nncf_node of type not from types list.
    """
    visited = {node_id: False for node_id in graph.get_all_node_ids()}
    partial_traverse_function = partial(traverse_function, type_check_fn=lambda x: x not in types, visited=visited)
    nncf_nodes = [nncf_node]
    if nncf_node.node_type not in types:
        nncf_nodes = graph.get_next_nodes(nncf_node)

    next_nodes = []
    for node in nncf_nodes:
        next_nodes.extend(graph.traverse_graph(node, partial_traverse_function))
    return next_nodes


def get_next_nodes_of_types(graph: NNCFGraph, nncf_node: NNCFNode, types: List[str]) -> List[NNCFNode]:
    """
    Looking for nodes with type from types list from `nncf_node` such that there is path
    from `nncf_node` to this node and on this path no node has one of types type.

    :param graph: Graph to work with.
    :param nncf_node: NNCFNode to start search.
    :param types: List of types to find.
    :return: List of next nodes of nncf_node with type from types list.
    """
    sources_types = types
    visited = {node_id: False for node_id in graph.get_all_node_ids()}
    partial_traverse_function = partial(traverse_function, type_check_fn=lambda x: x in sources_types, visited=visited)
    nncf_nodes = [nncf_node]
    if nncf_node.node_type in sources_types:
        nncf_nodes = graph.get_next_nodes(nncf_node)

    next_nodes = []
    for node in nncf_nodes:
        next_nodes.extend(graph.traverse_graph(node, partial_traverse_function))
    return next_nodes


def get_rounded_pruned_element_number(total: int, sparsity_rate: float, multiple_of: int = 8) -> int:
    """
    Calculates number of sparsified elements (approximately sparsity rate) from total such as
    number of remaining items will be multiple of some value.
    Always rounds number of remaining elements up.

    :param total: Total elements number.
    :param sparsity_rate: Proportion of zero elements in total.
    :param multiple_of: Number of remaining elements must be a multiple of `multiple_of`.
    :return: Number of elements to be zeroed.
    """
    remaining_elems = math.ceil((total - total * sparsity_rate) / multiple_of) * multiple_of
    return max(total - remaining_elems, 0)


def traverse_function(
    node: NNCFNode, output: List[NNCFNode], type_check_fn: Callable[[str], bool], visited: Dict[int, bool]
) -> Tuple[bool, List[NNCFNode]]:
    if visited[node.node_id]:
        return True, output
    visited[node.node_id] = True

    if not type_check_fn(node.node_type):
        return False, output

    output.append(node)
    return True, output


def get_last_nodes_of_type(graph: NNCFGraph, op_types: List[str]) -> List[NNCFNode]:
    """
    Looking for last node in graph with type in `op_types`.
    Last == layer with type in `op_types`, that there is a path from this layer to the model output
    such that there are no other operations with type in `op_types` on it.

    :param op_types: Types of modules to track.
    :param graph: Graph to work with.
    :return: List of all last pruned nodes.
    """
    graph_outputs = graph.get_output_nodes()  # NNCFNodes here

    visited = {node_id: False for node_id in graph.get_all_node_ids()}
    partial_traverse_function = partial(traverse_function, type_check_fn=lambda x: x in op_types, visited=visited)
    last_nodes_of_type = []
    for output in graph_outputs:
        last_nodes_of_type.extend(graph.traverse_graph(output, partial_traverse_function, False))

    return last_nodes_of_type


def get_previous_convs(
    graph: NNCFGraph, nncf_node: NNCFNode, pruning_types: List[str], stop_propagation_ops: List[str]
) -> List[NNCFNode]:
    """
    Returns source convolutions of the node.

    :return: List of source convolutions of node.
    """
    sources = get_sources_of_node(nncf_node, graph, pruning_types + stop_propagation_ops)
    sources = [source for source in sources if source.node_type in pruning_types]
    return sources


def get_prunable_layers_in_out_channels(graph: NNCFGraph) -> Tuple[Dict[NNCFNodeName, int], Dict[NNCFNodeName, int]]:
    """
    Collects the number of input and output channels for each prunable layer in the graph.

    :param graph: NNCFGraph
    :return Dictionary with the number of input channels to convolution and linear layers:
            {node_name: input_channels_num}
            Dictionary with the number of output channels from convolution and linear layers:
            {node_name: output_channels_num}
    """
    in_channels, out_channels = {}, {}
    for node in graph.get_all_nodes():
        if isinstance(node.layer_attributes, (ConvolutionLayerAttributes, LinearLayerAttributes)):
            name = node.node_name
            if name in in_channels and name in out_channels:
                continue
            in_channels[name] = get_input_channels(node)
            out_channels[name] = get_output_channels(node)
    return in_channels, out_channels


TObj = TypeVar("TObj", bound=type)


class PruningOperationsMetatypeRegistry(Registry):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self._op_name_to_op_class: Dict[str, type] = {}

    def register(self, name: Optional[str] = None) -> Callable[[TObj], TObj]:
        name_ = name
        super_register = super()._register

        def wrap(obj: TObj) -> TObj:
            cls_name = name_
            if cls_name is None:
                cls_name = obj.__name__
            super_register(obj, cls_name)
            op_names = obj.get_all_op_aliases()  # type: ignore
            for name in op_names:
                if name not in self._op_name_to_op_class:
                    self._op_name_to_op_class[name] = obj
                else:
                    assert (
                        self._op_name_to_op_class[name] == obj
                    ), "Inconsistent operator type registry - single patched op name maps to multiple metatypes!"
            return obj

        return wrap

    def get_operator_metatype_by_op_name(self, op_name: str) -> Optional[type]:
        if op_name in self._op_name_to_op_class:
            return self._op_name_to_op_class[op_name]
        return None


class PruningAnalysisReason(Enum):
    """
    Enum of possible pruning analysis decisions reasons.
    """

    IGNORED_SCOPE = "node in ignored scope"
    FIRST_CONV = "this scope is one of the first convolutions"
    LAST_CONV = "this scope is convolution with output which directly affects model output dimensions"
    GROUP_CONV = "this scope is grouped convolution"
    DOWNSAMPLE_CONV = "this scope is convolution with downsample"
    MODEL_ANALYSIS = "of model analysis"
    DIMENSION_MISMATCH = "of dimension mismatch"
    CLOSING_CONV_MISSING = "closing convolution missing"
    IN_GROUP_OF_UNPRUNABLE = "is in the group with non prunable layers"
    BATCHED_LINEAR = "linear node has batched dimension(s)"
    INCOMPATIBLE_DIMS_IN_CLUSTER = "channels in cluster nodes have different values"

    @classmethod
    def message(cls, node_name: str, decision: "PruningAnalysisDecision") -> str:
        """
        Returns the node pruning analysis decisions in a human-readable format.

        :param node_name: Name of given node.
        :param decision: Pruning analysis decision for given node.
        :return: Pruning analysis decision in a human-readable format.
        """
        prefix = f"ignored adding Weight Pruner in: {node_name}"
        reasons = decision.reasons
        if not reasons:
            return prefix
        # Filter messages
        if len(reasons) > 1 and cls.CLOSING_CONV_MISSING in reasons:
            reasons.remove(cls.CLOSING_CONV_MISSING)
        if len(reasons) == 1 and cls.IN_GROUP_OF_UNPRUNABLE in reasons:
            return ""
        return prefix + " because " + " and ".join([reason.value for reason in reasons])


class PruningAnalysisDecision:
    """
    Container for pruning analysis decisions. Contains decision which is boolean marker either
    node prunable or not (prunable if decision attribute is True) and
    pruning analysis reason in PruningAnalysisReason format. In case of positive
    decision (decision == True) possible reason will be ignored.
    """

    def __init__(
        self,
        decision: bool,
        possible_reasons: Optional[Union[List[PruningAnalysisReason], PruningAnalysisReason]] = None,
    ):
        self.decision = decision
        self._reasons: Optional[List[PruningAnalysisReason]] = None
        if possible_reasons is None:
            return
        if not isinstance(possible_reasons, list):
            possible_reasons = [possible_reasons]
        if not decision:
            self._reasons = possible_reasons

    def __repr__(self) -> str:
        representation = f"Prunable: {self.decision}"
        if not self.decision:
            representation += "; Reasons: " + str(self._reasons)
        return representation

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PruningAnalysisDecision):
            return False
        eq = self.decision == other.decision
        if self._reasons is None:
            return eq and other._reasons is None
        if other._reasons is None:
            return False
        return eq and set(self._reasons) == set(other._reasons)

    def __bool__(self) -> bool:
        return self.decision

    @property
    def reasons(self) -> Optional[List[PruningAnalysisReason]]:
        if self._reasons:
            return self._reasons.copy()
        return None

    def join(self, other: "PruningAnalysisDecision") -> "PruningAnalysisDecision":
        """
        Join two pruning analysis decisions about one NNCFNode.

        :param other: pruning analysis decision to join with.
        :return: Joint pruning analysis decision.
        """
        if self.decision and other.decision:
            return self

        reasons = []
        for decision in [self, other]:
            if decision.reasons:
                reasons.extend(decision.reasons)

        return PruningAnalysisDecision(False, reasons)


def is_prunable_depthwise_conv(node: NNCFNode) -> bool:
    # Only convolutions with in_channels == groups == out_channels are supported
    # by pruning algorithm. Depthwise convolutions support ticket: #68580
    return (
        isinstance(node.layer_attributes, ConvolutionLayerAttributes)
        and node.layer_attributes.groups == node.layer_attributes.in_channels
        and (node.layer_attributes.out_channels == node.layer_attributes.in_channels)
        and node.layer_attributes.in_channels > 1
    )


def is_conv_with_downsampling(node: NNCFNode) -> bool:
    layer_attrs = node.layer_attributes
    if isinstance(layer_attrs, ConvolutionLayerAttributes):
        return not np.all(np.array(layer_attrs.stride) == 1) and not layer_attrs.transpose
    return False


def get_input_masks(node: NNCFNode, graph: NNCFGraph) -> List[Optional[SymbolicMask]]:
    """
    Returns input masks for all inputs of given NNCFNode.

    :param node: Given NNCFNode.
    :param graph: Graph to work with.
    :return: Input masks.
    """
    retval = []
    input_masks = [input_edge.from_node.attributes["output_mask"] for input_edge in graph.get_input_edges(node)]
    for input_mask in input_masks:
        retval.append(input_mask[node.node_name] if isinstance(input_mask, dict) else input_mask)
    return retval


def get_input_channels(node: NNCFNode) -> int:
    """
    Returns count of input channels of an prunable node.

    :param node: Given prunable node.
    :return: Count of input channels of the given node.
    """
    layer_attrs = node.layer_attributes
    if isinstance(layer_attrs, ConvolutionLayerAttributes):
        return layer_attrs.in_channels
    if isinstance(layer_attrs, LinearLayerAttributes):
        return layer_attrs.in_features
    raise nncf.InternalError(f"Can't get count of input channels from node {node}")


def get_output_channels(node: NNCFNode) -> int:
    """
    Returns count of output channels of an prunable node.

    :param node: Given prunable node.
    :return: Count of output channels of the given node.
    """
    layer_attrs = node.layer_attributes
    if isinstance(layer_attrs, ConvolutionLayerAttributes):
        return layer_attrs.out_channels
    if isinstance(layer_attrs, LinearLayerAttributes):
        return layer_attrs.out_features
    raise nncf.InternalError(f"Can't get count of output channels from node {node}")


def identity_mask_propagation(node: NNCFNode, graph: NNCFGraph) -> None:
    """
    Propagates input mask through NNCFNode.

    :param node: Graph node to perform identity mask propagation on.
    :param graph: Graph to work with.
    """
    input_masks = get_input_masks(node, graph)
    if not input_masks:
        # In case for disconnected NNCFGraph
        input_masks = [None]
    assert len(input_masks) == 1

    node.attributes["output_mask"] = input_masks[0]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Set, Type

from nncf.common.graph import NNCFGraph
from nncf.common.pruning.operations import BasePruningOp
from nncf.common.pruning.symbolic_mask import SymbolicMask
from nncf.common.pruning.symbolic_mask import SymbolicMaskProcessor
from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.pruning.utils import PruningAnalysisDecision
from nncf.common.pruning.utils import PruningAnalysisReason
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.common.pruning.utils import get_input_channels
from nncf.common.pruning.utils import get_input_masks
from nncf.common.pruning.utils import get_output_channels
from nncf.common.pruning.utils import is_batched_linear
from nncf.common.pruning.utils import is_grouped_conv


class MaskPropagationAlgorithm:
    """
    Algorithm responsible for propagation masks across all nodes in the graph.
    Before call mask_propagation() you need set node.attributes['output_masks']
    for nodes that have masks already defined.
    """

    def __init__(
        self,
        graph: NNCFGraph,
        pruning_operator_metatypes: PruningOperationsMetatypeRegistry,
        tensor_processor: Optional[Type[NNCFPruningBaseTensorProcessor]] = None,
    ) -> None:
        """
        Initializes MaskPropagationAlgorithm.

        :param graph: Graph to work with.
        :param pruning_operator_metatypes: Registry with operation metatypes pruning algorithm is aware of, i.e.
               metatypes describing operations with common pruning mask application and propagation properties.
        :param tensor_processor: Framework-specific tensor processor.
        """
        self._graph = graph
        self._pruning_operator_metatypes = pruning_operator_metatypes
        self._tensor_processor = tensor_processor

    def get_meta_operation_by_type_name(self, type_name: str) -> Type[BasePruningOp]:
        """
        Returns class of metaop that corresponds to `type_name` type.

        :param type_name: Name of type of layer
        :return: Class of metaop that corresponds to `type_name` type.
        """
        cls = self._pruning_operator_metatypes.get_operator_metatype_by_op_name(type_name)
        if cls is None:
            cls = self._pruning_operator_metatypes.registry_dict["stop_propagation_ops"]
        return cls

    def mask_propagation(self) -> None:
        """
        Mask propagation in graph:
        to propagate masks run method mask_propagation (of metaop of current node) on all nodes in topological order.
        """
        for node in self._graph.topological_sort():
            cls = self.get_meta_operation_by_type_name(node.node_type)
            cls.mask_propagation(node, self._graph, self._tensor_processor)  # type: ignore

    def symbolic_mask_propagation(
        self, prunable_layers_types: List[str], can_prune_after_analysis: Dict[int, PruningAnalysisDecision]
    ) -> Dict[int, PruningAnalysisDecision]:
        """
        Check all nodes that were marked as prunable after the model analysis and compatibility check vs.
        pruning algo have a correct correspondent closing node on each path from self to outputs;
        the check entails verifying that every convolution prunable by the output channel dimension
        has a corresponding convolution that is prunable by its input channel dimension (output channel
        dimension equal to closing convolution input channel dimension) in every path from self to outputs.
        If the check fails, the entire groups containing such nodes will be marked as unprunable.
        If convolution symbolic mask mixes with other symbolic masks (by elementwise operation, for example)
        and mixing masks can't be mixed, all mask producers participated in this mixing will be marked as unprunable.
        If convolution output channel dimension reducing directly affect an output of the model -
        it will be marked as unprunable as well.


        :param prunable_layers_types: Types of operations with prunable filters.
        :param can_prune_after_analysis: Dict of node indices vs the decision made by previous steps;
            the decision is true only for the nodes that do not conflict with mask propagation and
            are supported by the NNCF pruning algorithm.
        :return: Dict of node indices vs the decision made by symbolic mask propagation algorithm.
        """

        can_be_closing_convs = self._get_can_closing_convs(prunable_layers_types)
        can_prune_by_dim: Dict[int, PruningAnalysisDecision] = {k: None for k in can_be_closing_convs}  # type: ignore
        for node in self._graph.topological_sort():
            if node.node_id in can_be_closing_convs and can_prune_after_analysis[node.node_id]:
                # Set output mask
                node.attributes["output_mask"] = SymbolicMask(get_output_channels(node), node.node_id)
            # Propagate masks
            cls = self.get_meta_operation_by_type_name(node.node_type)
            cls.mask_propagation(node, self._graph, SymbolicMaskProcessor)
            if node.node_id in can_be_closing_convs:
                # Check input mask producers out channel dimension
                input_masks = get_input_masks(node, self._graph)
                if any(input_masks):
                    assert len(input_masks) == 1
                    input_mask = input_masks[0]
                    assert isinstance(input_mask, SymbolicMask)

                    for producer in input_mask.mask_producers:
                        previously_dims_equal = (
                            True if can_prune_by_dim[producer.id] is None else can_prune_by_dim[producer.id]
                        )

                        is_dims_equal = get_input_channels(node) == input_mask.shape[0]
                        decision = bool(previously_dims_equal and is_dims_equal)
                        can_prune_by_dim[producer.id] = PruningAnalysisDecision(
                            decision, PruningAnalysisReason.DIMENSION_MISMATCH
                        )
        # Remove all convolutions with masks
        # that were propagated to output node
        for out_node in self._graph.get_output_nodes():
            for input_mask in get_input_masks(out_node, self._graph):
                if input_mask:
                    for producer in input_mask.mask_producers:
                        can_prune_by_dim[producer.id] = PruningAnalysisDecision(False, PruningAnalysisReason.LAST_CONV)
        # Update decision for nodes which
        # have no closing convolution
        convs_without_closing_conv: Dict[int, PruningAnalysisDecision] = {}
        for k, v in can_prune_by_dim.items():
            if v is None:
                convs_without_closing_conv[k] = PruningAnalysisDecision(
                    False, PruningAnalysisReason.CLOSING_CONV_MISSING
                )
        can_prune_by_dim.update(convs_without_closing_conv)

        # Clean nodes masks
        for node in self._graph.get_all_nodes():
            node.attributes["output_mask"] = None

        return can_prune_by_dim

    def _get_can_closing_convs(self, prunable_layers_types: List[str]) -> Set[int]:
        retval: Set[int] = set()
        for node in self._graph.get_all_nodes():
            if node.node_type in prunable_layers_types and not (
                is_grouped_conv(node) or is_batched_linear(node, self._graph)
            ):
                retval.add(node.node_id)
        return retval
